__version__ = "0.2.1"

#automatically imports all of the key submodule classes/functions into the PottsPlayground namespace
from PottsPlayground.Annealing import Anneal

import PottsPlayground.Schedules as Schedules
import PottsPlayground.Kernels as Kernels

import PottsPlayground.Tasks as Tasks #tasks relies on Kernels, so it needs to come after

#rebranding, without breaking existing code:
from PottsPlayground.Tasks.BaseTask import BaseTask as PottsModel
PottsModel.__name__ = 'PottsModel'
PottsModel.__module__ = 'PottsPlayground.PottsModel'


import numpy
import networkx as nx
from matplotlib import pyplot as plt
from PottsPlayground.Tasks.BaseTask import BaseTask
from PottsPlayground.Kernels import Identity as IdentityKernel

class GraphColoring(BaseTask):
	"""
	Models the Graph Node coloring problem as a Potts model, which is just about the simplest possible model.
	Rather than asking how many colors are needed to color a graph, which is the typical framing of graph coloring,
	the simple Potts model form asks if a coloring with a set number of colors is possible.
	"""

	def __init__(self, ncolors, p=None, nnodes=None, G=None, seed=None):
		"""
		Either construct a random Erdos-Renyi graph, or use a graph given by G.

		Parameters for either initialization:

		:param ncolors: How many colors are to be used.
		:type ncolors: int
		:param seed: Optional seed for random graph construction, for repeatable results.
		:type seed: int

		Parameters for coloring a random Erdos-Renyi graph:

		:param p: Edge probability for random graph construction (i.e. edge density)
		:type p: float
		:param nnodes: Number of graph nodes for random graph construction
		:param nnodes: int

		Parameters for coloring a given graph:
		
		:param G: The graph to be colored.
		:type G: undirected networkx graph
		"""
		BaseTask.__init__(self)
		# self.e_th = -100
		if G is not None:
			self.G = G
		else:
			self.G = nx.erdos_renyi_graph(nnodes, p, seed=seed)
		# self.Partitions2Nodes = list(self.G)
		# self.nnodes = len(self.Partitions2Nodes)
		self.ncolors = ncolors

		ik = lambda n: IdentityKernel(ncolors, n)
		
		#copy spins and edges from G to the BaseTask class:
		[self.AddSpins([ncolors], [node_name]) for node_name in self.G.nodes]
		[self.AddWeight(ik, i, j) for i, j in self.G.edges]
		

		# for i, node in enumerate(self.graph.nodes):
		# self.AddSpins(numpy.zeros([self.nnodes])+ncolors)	

		# for i, node1 in enumerate(self.Partitions2Nodes):
		# 	conn_nodes = [e[1] for e in self.G.edges(node1)]
		# 	for node2 in conn_nodes:
		# 		j = self.Partitions2Nodes.index(node2)
		# 		if i != j: #graph may have self-loops. we want to ignore them
		# 			self.AddKernel(lambda n: self.IdentityKernel(n), i, j)

		self.CompileKernels()

	def DisplayState(self, state):
		"""
		graphically represents the solution given by the supplied state.

		"""
		#map solution indices to colors for plotting:
		colors = ['tab:red', 'tab:green', 'tab:blue']
		coloring = [colors[c] for c in state]
		plt.figure(figsize=(7, 7))
		nx.draw_kamada_kawai(self.G, node_color=coloring)
		plt.show()


import numpy
import networkx as nx
from matplotlib import pyplot as plt
import PottsPlayground #import includes self. What will happen? No obvious issues.  Maybe the interpreter knows not to make this a circular thing.
from PottsPlayground.Tasks import BaseTask
from PottsPlayground.Tasks.GraphColoring import GraphColoring
import math
import time
import pickle
import copy
from collections import defaultdict

def GetCellPortNetName(cell, name):
	if cell.ports[name].net is not None:
		return cell.ports[name].net.name
	else:
		return None

def RecurseDownCarryChain(ctx, cell):
		#recursively finds the next cell in a carry chain, returning a list of all cells downwind of the given cell
		nn = GetCellPortNetName(cell, 'COUT')
		if nn is None:
			return [cell] #last cell in the carry chain
		net = ctx.nets[nn]
		return [cell] + RecurseDownCarryChain(ctx, net.users[0].cell)

def DffIncompatible(cell1, cell2):
	if (cell1.params['DFF_ENABLE']=='1') and (cell2.params['DFF_ENABLE']=='1'):
		if ((GetCellPortNetName(cell1, 'CEN') != GetCellPortNetName(cell2, 'CEN')) or
			(GetCellPortNetName(cell1, 'CLK') != GetCellPortNetName(cell2, 'CLK')) or
			(GetCellPortNetName(cell1, 'SR')  != GetCellPortNetName(cell2, 'SR'))  or
			(cell1.params['NEG_CLK'] != cell2.params['NEG_CLK'])):
				return True
	return False

def timing_paths(ctx):
	LC_in_ports = ["I0", "I1", "I2", "I3", "CIN"]

	#first, create directed graph where each node is a net,
	#and edges are non-dff cells between nets.
	G = nx.DiGraph()
	for key, net in ctx.nets:
		G.add_node(net.name, depth=1)

	for key, cell in ctx.cells:
		if cell.type != "ICESTORM_LC":
			continue #timing only for logic cells

		input_net_names = [(port, GetCellPortNetName(cell, port)) for port in LC_in_ports]
		input_net_names = [info for info in input_net_names if info[1] != None]

		output_net_names = []
		if cell.params['DFF_ENABLE'] == "0" and GetCellPortNetName(cell, 'O') is not None:
			output_net_names.append(('O', GetCellPortNetName(cell, 'O')))
		if GetCellPortNetName(cell, 'COUT') is not None:
			output_net_names.append(('COUT', GetCellPortNetName(cell, 'COUT')))

		for inp in input_net_names:
			for out in output_net_names:
				w = 0.1 if (inp[0] == 'CIN' and out == 'COUT') else 1.
				#carry connections are automatically very fast,
				#so should not be fully counted towards timing arc length
				G.add_edge(inp[1], out[1], weight=w)

	#delete two special nets that nextpnr uses:
	for special_node in ["$PACKER_VCC_NET", "$PACKER_GND_NET"]:
		if special_node in G:
			G.remove_node(special_node)

	# print(sorted(nx.simple_cycles(G)))
	assert len(sorted(nx.simple_cycles(G))) == 0

	#to figure out how important each net is to timing, traverse the graph
	#forwards and backwards, adding up the lengths each way.
	#after adding up, nodes in G_forward will know how the maximum depth
	#of nets before them, and G_reverse will know the maximum depth after.
	#adding the two gives the length of the longest pathway that passes through each net.
	#To make sure that each net really knows its maximum before or after length,
	#graph edges are deleted as they are followed and edges are only followed once all the 
	#preceding edges have been processed and deleted.
	G_forward = copy.deepcopy(G)
	G_reverse = copy.deepcopy(G)

	while (G_forward.number_of_edges() > 0):
		# print(G_forward.number_of_edges())
		for net in G_forward.nodes:
			if len(G_forward.in_edges(net)) == 0:
				to_remove = []
				for u, v, edge_data in G_forward.out_edges(net, data=True):
					G_forward.nodes[v]['depth'] = max(G_forward.nodes[v]['depth'], G_forward.nodes[u]['depth']+edge_data['weight'])
					# print(G.nodes[v]['depth'], G.nodes[u]['depth'], edge_data['weight'])
					to_remove.append((u, v))
				G_forward.remove_edges_from(to_remove)

	while (G_reverse.number_of_edges() > 0):
		# print(G_reverse.number_of_edges(), G_reverse.number_of_nodes())
		for net in G_reverse.nodes:
			if len(G_reverse.out_edges(net)) == 0:
				to_remove = []
				for u, v, edge_data in G_reverse.in_edges(net, data=True):
					G_reverse.nodes[u]['depth'] = max(G_reverse.nodes[u]['depth'], G_reverse.nodes[v]['depth']+edge_data['weight'])
					to_remove.append((u, v))
				G_reverse.remove_edges_from(to_remove)

	#combine forward and reverse counts:
	arc_lengths = defaultdict(lambda:1)
	for net in G.nodes:
		# print(G_reverse[net], G_forward[net])
		combined = G_reverse.nodes[net]['depth'] + G_forward.nodes[net]['depth'] - 1
		arc_lengths[net] = combined

	return arc_lengths

def label_carry_chains(ctx):
	for key, cell in ctx.cells:
		if cell.type == "ICESTORM_LC" and GetCellPortNetName(cell, 'COUT') is not None and GetCellPortNetName(cell, 'CIN') is None:
			#then this is the start of a chain!
			chain = RecurseDownCarryChain(ctx, cell)
			[chain_cell.setAttr("type", "CCchild") for chain_cell in chain] #re-type cells to indicate they are in a carry chain
			# [chain_cell.setAttr("CC", cell.name) for chain_cell in chain] 
			cell.setAttr("type", "CC") #re-type root cells

def label_lc_subtypes(ctx):
	for i, (key, cell) in enumerate(ctx.cells):
		if cell.type == "ICESTORM_LC":
			continue
		if "BEL" in cell.attrs:
			bel = cell.attrs["BEL"]
			loc = ctx.getBelLocation(bel)
			cell.setAttr("SubType", "%i"%loc.z)
		else:
			cell.setAttr("SubType", "%i"%i)

def gather_bel_types(ctx):
	BelTypes = defaultdict(list)
	# self.GetBelType = {} #replicate ctx functionality, so that I can easily, temporarily keep track of artificial bel type changes
	for bel in ctx.getBels():
		bel_type = ctx.getBelType(bel)
		# self.GetBelType[bel] = bel_type
		BelTypes[bel_type].append(bel)

	for i, bel in enumerate(BelTypes['ICESTORM_LC']):
		loc = ctx.getBelLocation(bel)
		BelTypes["LC%i"%(loc.z)].append(bel)
		# self.GetBelType[bel] = "LC%i"%loc.z

	BelTypes['CC'] = BelTypes['LC0'] #artificial mapping, so code below can apply more uniformly

	return BelTypes


class Ice40Placer(BaseTask.BaseTask):
	"""
	Potts Model that corresponds to placing logic elements in an Ice40 FPGA.  Ties in with the NextPNR tool;
	is not 100% functional, but can be used to place many FPGA designs.  Supports LUTs, carry chains, and IO.
	"""

	def __init__(self, ctx, cost_balancing=(15, 0.5, 1, 0), split_cells=True, verbose=False):
		"""
		Creates a model of FPGA placement given a NextPNR context.  The context is available when running a Python script inside of
		the NextPNR tool.  The context includes both information about the specific FPGA (availability and locations of physical Basic ELements i.e. BELs)
		and about the design that needs to be placed (logical BELs and their connectivity).  There are several different optimization objectives that
		can be balanced.  The first objective, 'exclusion', is mandatory, since it enforces that multiple logical BELs are not assigned to
		the same physical BEL.  'wirelen' tries to minimize overall distance between connected logical BELs; 'timing' weighs critical path connections
		more to keep the critical path as fast as possible; and 'jogs' tries to assign connected BELs to the same row or column so that 
		only a single horizontal or vertical routing channel is needed to connect them.

		:param ctx: A NextPNR context opbject.
		:param cost_balancing: Specifies relative weight of different optimization objectives. Objectives are (exclusion, wirelen, timing, jogs)
		:type cost_balancing: tuple
		:param split_cells: FPGA LUTs are grouped in blocks of eight; each of the eight have nearly similar connectivity.  If split_cells is True,
		each logical LUT is pre-constrained to one of the eight LUT positions.  This reduces the optimization space for faster results.
		:type split_cells: boolean
		:type verbose: boolean
		"""

		#======================================================construct friendlier formats of the design specifications
		#get lists of bel types in the architecture:
		BaseTask.BaseTask.__init__(self)
		exclusion_factor = cost_balancing[0]
		w_wirelen = cost_balancing[1]
		w_timing = cost_balancing[2]
		w_jogs = cost_balancing[3]

		net_timing_info = timing_paths(ctx)
		arc_lengths = [value for key, value in net_timing_info.items()]
		print("arc lengths", numpy.unique(arc_lengths, return_counts=True))
		longest_arc = numpy.max(arc_lengths)
		# print("Longest arc", longest_arc)

		label_carry_chains(ctx)
		
		if split_cells:
			label_lc_subtypes(ctx)

		self.BelTypes = gather_bel_types(ctx)
			

		#subdivide global buffer bels based on even-oddness, since they have slightly different features
		wire_names = [ctx.getBelPinWire(bel, "GLOBAL_BUFFER_OUTPUT") for bel in self.BelTypes["SB_GB"]]
		net_nums = [int(wire_name[-1]) for wire_name in wire_names]
		even_global_buffers = [i for i, netnum in enumerate(net_nums) if netnum%2 == 0]
		odd_global_buffers =  [i for i, netnum in enumerate(net_nums) if netnum%2 == 1]

		#get a sub-list of logic blocks that are not at the top,
		#i.e. ones that are not limited to length-8 carry chains
		bel_loc = [ctx.getBelLocation(bel) for bel in self.BelTypes['CC']]
		bel_y = [loc.y for loc in bel_loc]
		ymax = numpy.max(bel_y)
		constrain2lowerY = [i for i, y in enumerate(bel_y) if y != ymax]
		
		#======================================================= partition construction
		self.used_bels = []
		for key, cell in ctx.cells:
			if cell.type == "CCchild":
				continue #these are not explicitly included in the optimization

			self.AddSpins([len(self.BelTypes[cell.type])], [cell.name])

			# not all global buffers can drive all signal types.
			# Must restrict certain global buffer cells to a subset of physical global buffers:
			if cell.type == "SB_GB":
				users = cell.ports["GLOBAL_BUFFER_OUTPUT"].net.users
				user_ports = [user.port for user in users]
				if "CEN" in user_ports:
					self.PinSpin(cell.name, odd_global_buffers)

			if cell.type == "CC":
				self.PinSpin(cell.name, constrain2lowerY)

			#if a cell has already been fixed for some reason (like IO),
			#fix it in the Potts model here too:
			if "BEL" in cell.attrs:
				bel = cell.attrs["BEL"]
				self.used_bels.append(bel)
				bel_position = self.BelTypes[cell.type].index(bel)
				self.PinSpin(cell.name, [bel_position])


		
			# print("")
			# for thing in dir(cell):
				# print(thing)
			# for k, v in cell.ports:
				# print(k, v)
				# for thing in v.net.users:
					# print(thing.port)
					# for thing2 in dir(thing.port):
						# print(thing2)
			# for k, v in cell.attrs:
				# print(k, v)

		# exit()

		
		# ============================================================================= kernel map construction
		


		#I'm not quite sure what I was doing here.  It's not critical to functioning, so let's ignore it for now
		# total_weight_strength = numpy.zeros([nPartitions])
		# for i, cell1_name in enumerate(self.Partitions2CellNames):
		# 	for j, cell2_name in enumerate(self.Partitions2CellNames):
		# 		if G.has_edge(cell1_name, cell2_name):
		# 			data = G[cell1_name][cell2_name]
		# 			nEndpoints = data["w"]
		# 			total_weight_strength[i] = total_weight_strength[i] + w_wirelen/nEndpoints
		# 			d2u = data["d2u"]
		# 			if "ArcsAfter" in data and "ArcsBefore" in data:
		# 				narcs = data["ArcsAfter"]+data["ArcsBefore"]+1
		# 				total_weight_strength[i] = total_weight_strength[i] + w_timing*2**(narcs-maxnarc)
		# 				total_weight_strength[i] = total_weight_strength[i] + w_jogs

		# print(total_weight_strength)

		lc_check_types = ["LC%i"%i for i in range(8)] + ["CC", "ICESTORM_LC"]

		#set up ntiles variables, which tells if a cell occupies 1 or more than 1 logic tiles (if it is a carry chain)
		ntiles = {}
		for key, cell in ctx.cells:
			if cell.type == "CC":
				ntiles[cell.name] = math.ceil(len(RecurseDownCarryChain(ctx, cell))/8)
			else:
				ntiles[cell.name] = 1

		for key1, cell1 in ctx.cells:
			if cell1.type == "CCchild":
				continue #these are not explicitly included in the optimization
			for key2, cell2 in ctx.cells:
				if cell2.type == "CCchild":
					continue #these are not explicitly included in the optimization
				if cell1.name == cell2.name:
					continue

				#there are two types of hard constraints, bel exclusion and tile exclusion.
				#Bel exclusion means that two cells have the same type, and therefore cannot
				#be in the same place.  Simple.
				#Tile exclusion means that two cells cannot even be in the same tile,
				#due to various factors.  A carry chain in one tile might even exclude
				#other cells in the tile above it, if the chain is longer than 8.

				
				tile_exclude = False
				
				#two logic cells with different D flip flop configurations cannot share the same logic tile
				if cell1.type in lc_check_types and cell2.type in lc_check_types:
					tile_exclude = DffIncompatible(cell1, cell2)

				#if a cell is a carry chain, it may span across more than one tile,
				#in which case we should have a multi-tile exclusion.
				#we must also account for when two multi-length carry chains exclude each other.
				ntiles1 = ntiles[cell1.name]
				ntiles2 = ntiles[cell2.name]

				#create a structural penalty to prevent cells from sharing the same FPGA resources
				#(applies to the mutual exclusion of carry chains and logic cells too)
				if (cell1.type == "CC" and cell2.type in lc_check_types):
					tile_exclude = True
				if (cell2.type == "CC" and cell1.type in lc_check_types):
					tile_exclude = True

				wgt = 1*exclusion_factor
				if tile_exclude:
					self.AddKernel(lambda n: self.TileExclusion(ctx, cell1.type, cell2.type, ntiles1, ntiles2, n), cell1.name, cell2.name, weight=wgt)

				elif cell1.type == cell2.type:
					self.AddKernel(lambda n: self.BelExclusion(n), cell1.name, cell2.name, weight=wgt)


		#new loop through nets, to add net constraints.
		#actually, could do this later.  It's a secondary goal...
				#determine the wiring connectivity between the two cells,
				# connected = False
				# if G.has_edge(cell1_name, cell2_name):
				# 	connected = True #the heavy lifting and case handling has been done elsewhere, in the construction of G
				# 	data = G[cell1_name][cell2_name]
				# 	nEndpoints = data["w"]
				# 	d2u = data["d2u"]
				# 	if "ArcsAfter" in data and "ArcsBefore" in data:
				# 		narcs = data["ArcsAfter"]+data["ArcsBefore"]+1

				# #a geometric mean of how constrained each cell is determines how much constraint weight there should be
				# wgt = 1+total_weight_strength[i]*total_weight_strength[j]/(total_weight_strength[i] + total_weight_strength[j])

				# # =====================================================main if-else to decide which kernel, if any, relates each pair of cells
				# if i==j:
				# 	#the assigned position of one cell should not feed back upon itself
				# 	continue

				# if connected and d2u:
				# 	self.AddKernel(lambda n: self.NoJogsKernel(ctx, type1, type2, n), i, j, weight=w_jogs)
				# 	self.AddKernel(lambda n: self.TimingKernel(ctx, type1, type2, n), i, j, weight=w_timing*2**(narcs-maxnarc))

				# elif connected:
				# 	self.AddKernel(lambda n: self.WirelenKernel(ctx, type1, type2, n), i, j, weight=w_wirelen/nEndpoints)


		self.CompileKernels() #a baseTask function


		#revert altered types:
		for key, cell in ctx.cells:
			if cell.type.startswith("CC") or cell.type.startswith("LC"):
				cell.setAttr("type", "ICESTORM_LC")
			# print(cell.type)

	

	def SetResultInContext(self, ctx, state):
		"""
		Takes the final state of an annealing run and configures that result into the nextpnr context.
		The context is not returned but afterwards should contain added constraints corresponding to the given state.

		:param ctx: NextPNR context corresponding to the PlacerTask.
		:param state: A Potts model state.
		:type state: 1-D Numpy int array
		:return: Number of conflicts, i.e. number of BELs assigned to incompatible physical locations.  Should be 0 for a successful run.
		:rtype: int
		"""
		nConflicts = 0
		for i, q in enumerate(state):
			# if i > 5:
				# return
			cell_name = self.Partitions2CellNames[i]
			cell = ctx.cells[cell_name]
			if "BEL" in cell.attrs:
				#this cell was previously constrained, so we don't set it:
				continue
			cell_type = self.GetCellType[cell_name]
			if cell_type == "CC":
				chain = [ctx.cells[cn] for cn in self.chains[cell_name]]
				root_bel = self.BelTypes['LC0'][q]
				print("constraining length", len(chain), "carry chain with cell", cell_name,  "rooted at bel", root_bel)
				for j, cell in enumerate(chain):
					LCtype = "LC%i"%(j%8)
					bel = self.BelTypes[LCtype][q+int(j/8)]#move to the next tile for each 8 logic cells
					if bel not in self.used_bels:
						cell.setAttr("BEL", bel)
						# ctx.bindBel(bel, cell, STRENGTH_STRONG)
						self.used_bels.append(bel)
					else:
						print("Error, cell %s cannot be assigned to bel %s which is already assigned"%(cell_name, bel))
						nConflicts = nConflicts + 1
			else: #all other things that are not carry chains
				bel = self.BelTypes[cell_type][q]
				# print(cell.name, bel)
				# if (ctx.checkBelAvail(bel)):
				if bel not in self.used_bels:
					# ctx.bindBel(bel, cell, strength)
					cell.setAttr("BEL", bel)
					# ctx.bindBel(bel, cell, STRENGTH_STRONG)
					self.used_bels.append(bel)
					# cell.attrs["BEL"] = bel
				else:
					print("Error, cell %s cannot be assigned to bel %s which is already assigned"%(cell_name, bel))
					nConflicts = nConflicts + 1

		return nConflicts

	# =============================================================================================== kernel generators
	# =================================================================================================================
	def get_type_locs(self, ctx, type1):
		locs1 = numpy.zeros([len(self.BelTypes[type1]), 2])
		for i, bel1 in enumerate(self.BelTypes[type1]):
			loc1 = ctx.getBelLocation(bel1)
			locs1[i, 0] = loc1.x
			locs1[i, 1] = loc1.y
		return locs1

	def NoJogsKernel(self, ctx, type1, type2, n=False):
		if type1.startswith("LC") or type1 == "CC":
			type1 = "LC0"
		if type2.startswith("LC") or type2 == "CC":
			type2 = "LC0" #has to be set to a specific type, so that it can find the right bel distances

		name = "NoJogs between types " + type1 + " and " + type2

		if n:
			return name
		elif name in self.KernelDict:
			#prevent redundant construction (possibly caused by constructing compound kernels)
			return self.KernelList[self.KernelDict[name]]

		locs1 = self.get_type_locs(ctx, type1)
		locs2 = self.get_type_locs(ctx, type2)
		locs1 = numpy.expand_dims(locs1, 1)
		locs2 = numpy.expand_dims(locs2, 0)

		dx = numpy.abs(locs1[:,:,0]-locs2[:,:,0])
		dy = numpy.abs(locs1[:,:,1]-locs2[:,:,1])

		local_conn = (dx <= 1)*(dy<=1)*1
		dx_conn = (dx > 1)*(dy == 0) * 1
		dy_conn = (dy > 1)*(dx == 0) * 1
		distant_conn = (local_conn == 0)*(dx_conn == 0)*(dy_conn == 0)*(10+(dx+dy))

		return distant_conn

	def TimingKernel(self, ctx, type1, type2, n=False):
		#LC and carry chain types all use the same resource type, and have the same kernel size, so they share the same wirelen kernels
		if type1.startswith("LC") or type1 == "CC":
			type1 = "LC0"
		if type2.startswith("LC") or type2 == "CC":
			type2 = "LC0" #has to be set to a specific type, so that it can find the right bel distances

		name = "Timing between types " + type1 + " and " + type2

		if n:
			return name
		elif name in self.KernelDict:
			#prevent redundant construction (possibly caused by constructing compound kernels)
			return self.KernelList[self.KernelDict[name]]

		locs1 = self.get_type_locs(ctx, type1)
		locs2 = self.get_type_locs(ctx, type2)
		locs1 = numpy.expand_dims(locs1, 1)
		locs2 = numpy.expand_dims(locs2, 0)

		dx = numpy.abs(locs1[:,:,0]-locs2[:,:,0])
		dy = numpy.abs(locs1[:,:,1]-locs2[:,:,1])

		local_conn = (dx <= 1)*(dy<=1)*0.6
		dx_conn = (dx > 1)*(dy == 0) * (dx + 5)
		dy_conn = (dy > 1)*(dx == 0) * (dy + 5)
		distant_conn = (local_conn == 0)*(dx_conn == 0)*(dy_conn == 0)*3*(3+dx+dy)

		return dx_conn + dy_conn + distant_conn

	def WirelenKernel(self, ctx, type1, type2, n=False):
		#LC and carry chain type all use the same resource type, and have the same kernel size, so they share the same wirelen kernels
		if type1.startswith("LC") or type1 == "CC":
			type1 = "LC0"
		if type2.startswith("LC") or type2 == "CC":
			type2 = "LC0" #has to be set to a specific type, so that it can find the right bel distances

		name = "Wirelength between types " + type1 + " and " + type2

		if n:
			return name
		elif name in self.KernelDict:
			#prevent redundant construction (possibly caused by constructing compound kernels)
			return self.KernelList[self.KernelDict[name]]

		locs1 = self.get_type_locs(ctx, type1)
		locs2 = self.get_type_locs(ctx, type2)
		locs1 = numpy.expand_dims(locs1, 1)
		locs2 = numpy.expand_dims(locs2, 0)

		dx = numpy.abs(locs1[:,:,0]-locs2[:,:,0])
		dy = numpy.abs(locs1[:,:,1]-locs2[:,:,1])

		return dx + dy

	def TileExclusion(self, ctx, type1, type2, ntiles2, ntiles1, n=False):
		#LC and carry chain types all use the same resource type, and have the same kernel size, so they share the same wirelen kernels
		if type1.startswith("LC") or type1 == "CC":
			type1 = "LC0"
		if type2.startswith("LC") or type2 == "CC":
			type2 = "LC0" #has to be set to a specific type, so that it can find the right bel distances

		if n:
			return "TileExclusion-%s-%s-%i-%i"%(type1, type2, ntiles1, ntiles2)

		#numpy-vectorized kernel construction for improved speed
		locs1 = self.get_type_locs(ctx, type1)
		locs2 = self.get_type_locs(ctx, type2)
		locs1 = numpy.expand_dims(locs1, 1)
		locs2 = numpy.expand_dims(locs2, 0)

		loc1y = locs1[:,:,1]
		loc1x = locs1[:,:,0]
		loc2y = locs2[:,:,1]
		loc2x = locs2[:,:,0]
		return (loc1y-ntiles1 < loc2y)*(loc2y < loc1y+ntiles2)*(loc1x == loc2x)

	def BelExclusion(self, n=False):
		if n:
			return "BelExclusion"

		qMax = numpy.max([len(bels) for bels in self.BelTypes])
		return numpy.eye(qMax)


	def examine_path(self, path_cell_names):

		for i in range(1, len(path_cell_names)):
			cell1 = path_cell_names[i-1]
			cell2 = path_cell_names[i]
			if cell1 in self.G and cell2 in self.G:
				print(self.G[cell1][cell2])


if __name__ == '__main__':
	#alternate version that makes direct edits to the ctx instead of creating a parallel accounting structure:
	for key, cell in ctx.cells:
		if GetCellPortNetName(cell, 'COUT') is not None and GetCellPortNetName(cell, 'CIN') is None:
			#then this is the start of a chain!
			chain = RecurseDownCarryChain(ctx, cell)
			[chain_cell.setAttr("type", "CC") for chain_cell in chain] #re-type cells to indicate they are in a carry chain
			ctx.setAttr("type", "CCroot") #re-type root cells

	exit()

	placer = Ice40Placer(ctx, cost_balancing=(15, 0.5, 1, 0.1))
	tstart = time.perf_counter()
	results = Annealing.Anneal(vars(placer), placer.defaultTemp(niters=5e6, tmax=12), OptsPerThrd=1, TakeAllOptions=True, backend="PottsJit", substrate="CPU", nReplicates=1, nWorkers=1, nReports=1)
	ttot = time.perf_counter() - tstart
	print("Annealing time is %.2f seconds"%ttot) 

	placer.SetResultInContext(ctx, results['MinStates'][-1,:])


import PottsPlayground
import random
import numpy
import scipy.stats

print("Testing: PottsPlayground.Anneal")

#====================================================================================================
#test set 1: check that minimum TSP energy can be found:
PottsTask = PottsPlayground.Tasks.TravelingSalesman(ncities=5, ConstraintFactor=1.5)

valid_energies, invalid_energies = PottsTask.EnergyBands()
optE = numpy.min(valid_energies)

solver_temp = PottsPlayground.Schedules.SawtoothTempLog2Space(temp=1., MinTemp=0.05, nTeeth=3, niters=1e4)

#randomly check a number of operating combinations, 
for i in range(25):
	args = {}
	args['model'] = random.choice(["Potts", "PottsPrecompute", "Tsp"])
	args['device'] = random.choice(["CPU", "GPU"])
	args['algo'] = random.choice(["Simple", "BirdsEye", "OptionsActions"])
	args['nReplicates'] = random.choice([1,3,41])
	args['nWorkers'] = random.choice([1,3,41])
	args['nActions'] = random.choice([1,2,3]) #only applies if algo=OptionsActions
	args['nOptions'] = random.choice([1,5,1000]) #only applies if algo=OptionsActions

	results = PottsPlayground.Anneal(PottsTask, solver_temp, **args)

	if results["MinEnergies"][-1] > optE*1.001 or results["MinEnergies"][-1] < optE*0.999:
		print("FAILURE: did not return expected optimal result, settings =")
		print(args)
		exit()


#=======================================================================================================
#test 2: that initial conditions set correctly:
print("testing setting initial conditions, expect two warnings:")
PottsTask = PottsPlayground.Tasks.TravelingSalesman(ncities=5, ConstraintFactor=1.5)
temp = PottsPlayground.Schedules.LinearTemp(temp=1, niters=0)
for device in ['CPU', 'GPU']:
	#test single broadcast initial condition
	IC = numpy.random.randint(0, high=5, size=[5], dtype=numpy.int32)
	result = PottsPlayground.Anneal(PottsTask, temp, device=device, nReplicates=4, IC=IC)
	for i in range(4):
		assert (result['AllStates'][-1,i,:] == IC).all()
	#test unique initial conditions
	IC = numpy.random.randint(0, high=5, size=[4,5], dtype=numpy.int32)
	result = PottsPlayground.Anneal(PottsTask, temp, device=device, nReplicates=4, IC=IC)
	assert (result['AllStates'][-1,:,:] == IC).all()
	#test that the initial conditions can be left empty using the None keyword:
	PottsPlayground.Anneal(PottsTask, temp, device=device, nReplicates=4, IC=None)
	#test that the program does not segfault when the initial conditions are wrong dimensions:
	PottsPlayground.Anneal(PottsTask, temp, device=device, nReplicates=3, IC=IC)

#=========================================================================================================
#test set 3: make sure that average sampled energy is right (verify boltzmann distribution)
# PottsTask = PottsPlayground.Tasks.TravelingSalesman(ncities=3, ConstraintFactor=1.5)
# IsingTask = PottsPlayground.Tasks.Binarized(PottsTask, ConstraintFactor=1.5)

# nSamples = 100
# ref_temp = 2.
# temp = PottsPlayground.Schedules.constTemp(niters=1e4, temp=ref_temp)
# results = PottsPlayground.Anneal(IsingTask, temp, nReports=nSamples)
# ref_samples = results["AllEnergies"]


# #randomly check a number of operating combinations, 
# for i in range(25):
# 	args = {}
# 	args['model'] = random.choice(["Ising", "Potts", "PottsPrecompute", "PottsPrecomputePE"])
# 	args['device'] = random.choice(["CPU", "GPU"])
# 	if args['device'] == "CPU":
# 		args['algo'] = random.choice(["Simple", "BirdsEye", "OptionsActions"])
# 	else:
# 		args['algo'] = random.choice(["OptionsActions"])
# 		args['nWorkers'] = random.choice([1,4,16])
# 	if args['algo'] == "OptionsActions":
# 		args['nOptions'] = random.choice([1,1000])

# 	#have to set the right temperature, depending on the algorithm
# 	if (args['algo'] == "BirdsEye" and args["model"] == "PottsPrecomputePE"):
# 		temp = PottsPlayground.Schedules.constTemp(niters=1e4, temp=ref_temp)
# 	elif (args['algo'] == "BirdsEye"):
# 		temp = PottsPlayground.Schedules.constTemp(niters=1e4, temp=2*ref_temp)
# 	elif (args['algo'] == "OptionsActions" and args['nOptions'] == 1000 and args["model"] == "PottsPrecomputePE"):
# 		temp = PottsPlayground.Schedules.constTemp(niters=1e4, temp=2*ref_temp)
# 	elif (args['algo'] == "OptionsActions" and args['nOptions'] == 1000):
# 		temp = PottsPlayground.Schedules.constTemp(niters=1e4, temp=4*ref_temp)
# 	elif (args['algo'] == "OptionsActions"):
# 		temp = PottsPlayground.Schedules.constTemp(niters=1e4, temp=ref_temp)
# 	else:
# 		temp = PottsPlayground.Schedules.constTemp(niters=1e4, temp=ref_temp)

# 	results = PottsPlayground.Anneal(IsingTask, temp, nReports=nSamples, **args)

# 	if (args['algo'] == "BirdsEye" or ("nOptions" in args and args["nOptions"] == 1000)):
# 		#these have a continuous time interpretation, 
# 		#and thus must be scaled by the dwell time:
# 		dwell = results["DwellTimes"]
# 		norm_dwell = dwell/numpy.mean(dwell)
# 		samples = results["AllEnergies"]*norm_dwell
# 	else:
# 		samples = results["AllEnergies"]

	
# 	#two-sample t-test that this average energy and reference avg energy come from the same distribution:
# 	p = scipy.stats.ttest_ind(a=samples, b=ref_samples, equal_var=True)
# 	# print(p)
# 	# e_avg = numpy.mean(samples)
# 	# e_var = numpy.var(samples)
# 	# T = (e_avg - ising_e_avg)/(e_var/nSamples+ising_e_var/nSamples)**0.5

# 	# sp = ((nSamples*ising_e_var+nSamples*e_var)/(2*nSamples))**0.5
# 	# p = (ising_e_avg-e_avg)*(nSamples/2.)**0.5 / sp
# 	if (p.pvalue < 0.05):
# 		print("The average sampled energy is wrong! p=%.3f"%p.pvalue)
# 		print("Expected Eavg: %.3f Actual Eavg: %.3f"%(numpy.mean(ref_samples), numpy.mean(samples)))
# 		print(args)
# 		# exit()


import networkx as nx
from PottsPlayground.Tasks import BaseTask
from PottsPlayground.Kernels import BinaryQuadratic as BQK
import numpy

# import dwave.embedding
# import dimod
# import minorminer


class MinorEmbedded(BaseTask.BaseTask):
	"""
	Uses utilities from d-wave ocean libraries to minor embed and map from a q=2 potts model with arbitrary connectivity
	to a fixed-hardware topology ising model. 
	"""

	def __init__(self, task, target_graph, ConstraintFactor=1):
		"""
		Creates the minor-embedded task.  May take a long time, and success is not guarenteed, since minor-embedding
		is itself an NP hard problem.

		:param task: An Ising-compatible Potts task (i.e. a Potts task with q=2 for all Spins)
		:param target_graph: A NetworkX graph of the allowed Ising model connectivity.
		:param ConstraintFactor: Energy penalty for disagreements within each physical chain representing a single logical spin.
		:type ConstraintFactor: float
		"""
		BaseTask.BaseTask.__init__(self)
		print("Finding embedding.")#this can easily take a very long time, so it is worthwhile to let user know what is up
		

		# self.source_label_mapping = {original_label: idx for idx, original_label in enumerate(G.nodes())}
		# self.source_embeddable_graph = nx.relabel_nodes(task.graph, self.source_label_mapping)

		embedding = minorminer.find_embedding(task.graph, target_graph)
		self.embedding = embedding
		self.target = target_graph
		self.e_th = -1e9
		self.task = task

		if len(embedding) == 0:
			print("Error: no embedding found.")
			return
		else:
			target_nodes = numpy.sum([len(embedding[src]) for src in embedding])
			source_nodes = numpy.sum([1 for src in embedding])
			available_nodes = target_graph.number_of_nodes()
			print("Minor embedding maps %i source nodes to %i of %i nodes available in the target graph"%(source_nodes, target_nodes, available_nodes))
		
		#convert source q=2 potts graph into format needed by dwave.embedding.embed_ising
		source_h = {}
		source_J = {}
		c = 0

		#assume/work with binary quadratic form, with variables having values of 0 or 1.
		for i, spin in enumerate(task.ListSpins()):
			h0 = task.biases[i,0]
			h1 = task.biases[i,1]	
			#treat h0 as a "background" contribution to the energy,
			#that should be moved from the local bias and rolled into a global offset
			h = h1-h0
			c = c + h0
			source_h[spin] = h
			for spin1, spin2, edge_data in task.graph.out_edges(spin, data=True):
				if (spin2, spin1) in source_J:
					continue #checks if the transpose weight has already been processed and added to the edge dictionary
				k = edge_data['kIndex']
				k = task.KernelList[k]*edge_data['weight']
				assert (k.shape == (2,2) and k[0,0] == 0. and k[0,1] == 0. and k[1,0] == 0.) #enforce that source is in binary quadratic form
				wij = k[1,1]
				source_J[(spin1, spin2)] = wij

		

		dwave_src_model = dimod.BinaryQuadraticModel(source_h, source_J, c, 'BINARY')  #the init func here is shit and you can't use keyword args
		dwave_embedded_model = dwave.embedding.embed_bqm(dwave_src_model, embedding, target_graph, chain_strength=ConstraintFactor)

		#convert dwave's returned ising format back into a PottsPlayground task:
		emb_h = dwave_embedded_model.linear
		emb_J = dwave_embedded_model.quadratic
		emb_c = dwave_embedded_model.offset

		qSizes = numpy.array([2]*target_nodes)
		self.SetPartitions(qSizes)
		self.nnodes = target_nodes

		#map between the target graph node numbers, and the indices of the MinorEmbedded PottsPlayground task
		self.Indx2TgtNode = [n for n in emb_h]
		self.TgtNode2Indx = {tgt:i for i, tgt in enumerate(self.Indx2TgtNode)}

		# self.biases = numpy.add(self.biases, emb_c/self.nnodes, dtype="float32") 
		for i in range(self.nnodes):
			TgtNode = self.Indx2TgtNode[i]
			self.AddBias(i, [emb_c/self.nnodes, emb_c/self.nnodes]) #redistribute the offset across all the nodes
			self.AddBias(i, [0, emb_h[TgtNode]])
			# self.biases[i,1] = self.biases[i,1] + emb_h[TgtNode]

		# self.InitKernelManager()

		for edge in emb_J:
			i = self.TgtNode2Indx[edge[0]]
			j = self.TgtNode2Indx[edge[1]]
			w = emb_J[edge]
			self.AddWeight(BQK, i, j, weight=w)
			# self.AddKernel(lambda n: self.BQMKernel(n), j, i, weight=w)

		self.CompileKernels()



	# def BQMKernel(self, n=False):
	# 	if n:
	# 		return "ik"
	# 	k = numpy.zeros([2,2])
	# 	k[1,1] = 1
	# 	return k

	def StateToMinor(self, state):
		"""
		Converts a state in the native Ising format to a state in the embedded format.
		This is useful for creating an initial condition for the embedded model that corresponds to 
		an equal initial condition in the original model.
		
		:param state: A vector of spin values, corresponding to spins in the source/logical Ising model.
		:return: A vector of spin values, corresponding to spins in the target/physical Ising model.
		"""
		minor_state = numpy.zeros([self.nnodes], dtype="int32")
		for i_src, value in enumerate(state):
			tgt_nodes = self.embedding[i_src]
			for tgt_node in tgt_nodes:
				i = self.TgtNode2Indx[tgt_node]
				minor_state[i] = value
		return minor_state


	def FuzzyMinorToState(self, minor_state):
		"""
		Uses majority-vote to collapse the minor-embedded representation back into the source ising model.

		:param minor_state: A vector of spin values, corresponding to spins in the target/physical Ising model.
		:return: A vector of spin values, corresponding to spins in the source/logical Ising model.
		"""
		state = numpy.zeros([self.task.nnodes], dtype="int32")
		for i in range(self.task.nnodes):
			minor_nodes = self.embedding[i]
			minor_node_values = [minor_state[self.TgtNode2Indx[ii]] for ii in minor_nodes]
			state[i] = 1*(numpy.mean(minor_node_values) > 0.5)
		return state

	def IsValidSemantics(self, minor_state):
		for i in range(self.task.nnodes):
			minor_nodes = self.embedding[i]
			minor_node_values = [minor_state[self.TgtNode2Indx[ii]] for ii in minor_nodes]
			chain_avg = numpy.mean(minor_node_values)
			if chain_avg > 0.0001 and chain_avg < 0.9999:
				return False
		#if here, state is valid from perspective of the minor embedding.
		#Still need to check if it is valid w.r.t. the source task:
		return self.task.IsValidSemantics(self.FuzzyMinorToState(minor_state))




import PottsPlayground
import random
import numpy

print("Testing: PottsPlayground.Tasks.TravelingSalesman")

PottsTask = PottsPlayground.Tasks.TravelingSalesman(ncities=5, ConstraintFactor=10)
valid_energies, invalid_energies = PottsTask.EnergyBands()

exp_val = 5*4*3*2*1
if len(valid_energies) != exp_val:
	print("FAILURE: expected %i valid states, but there are %i"%(exp_val, len(valid_energies)))
	exit()

if numpy.max(valid_energies) > numpy.min(invalid_energies):
	print("FAILURE: invalid energies should all be greater than valid energies")
	exit()

dists = numpy.array(
	[[0,1,10,1],
	 [1,0,1,10],
	 [10,1,0,1],
	 [1,10,1,0]]
	)
PottsTask = PottsPlayground.Tasks.TravelingSalesman(costs=dists, ConstraintFactor=10)
valid_energies, invalid_energies = PottsTask.EnergyBands()
if numpy.min(valid_energies) != 4:
	print("FAILURE: incorrect min energy")
	exit()
if numpy.max(valid_energies) != 22:
	print("FAILURE: incorrect max energy")
	exit()




import numpy
from PottsPlayground.Tasks import BaseTask
from matplotlib import pyplot as plt


class TravelingSalesman(BaseTask.BaseTask):
	"""
	Potts model representation of the Traveling Salesman Problem (TSP).
	"""

	def __init__(self, ncities=None, costs=None, ConstraintFactor=1, seed=None):
		"""
		Two initializations are possible.  Either a regular euclidean traveling salesman on a plane
		is randomly generated, bounded in the coordinate box (0,1)x(0,1); or a matrix of segment costs is passed.

		For generating a regular euclidean TSP instance:

		:param ncities: The number of cities to randomly generate.
		:type ncities: int
		:param seed: Optional, define a seed to make the process repeatable.

		For using an arbitrary set of costs:

		:param costs: Matrix where each element i,j is the cost to travel from i to j.
		:type costs: NxN Numpy float
		"""

		#set up task-specific variables =========================================
		BaseTask.BaseTask.__init__(self)
		if seed is not None:
			numpy.random.seed(seed)

		if costs is None:
			#set up distances here, so that it only needs to be set up once:
			self.coords = numpy.random.random([2,ncities])
			self.distances = numpy.zeros([ncities,ncities], dtype="float32")
			for i in range(ncities):
				for j in range(i+1,ncities):
					d = ((self.coords[0,i]-self.coords[0,j])**2+(self.coords[1,i]-self.coords[1,j])**2)**0.5
					self.distances[i,j] = d
					self.distances[j,i] = d
		else:
			assert costs.shape[0] == costs.shape[1]
			self.distances = costs #allow an external distance/cost matrix to be used
			ncities = costs.shape[0]

		self.e_th = -1e14
		self.ncities = ncities
		self.nnodes = ncities
		self.ConstraintFactor = ConstraintFactor

		#create actual Potts formulation of the task =============================
		#set up partitioning - BaseTask function
		self.SetPartitions(numpy.zeros([ncities]) + ncities)

		for i in range(self.ncities):
			for j in range(self.ncities):
				if (i==j):
					continue
				self.AddKernel(lambda n: self.constrain(n), i, j, weight=1)

		for i in range(self.ncities):
			self.AddKernel(lambda n: self.distance(n), i, (i+1)%self.ncities, weight=1)
			self.AddKernel(lambda n: self.distance(n), (i+1)%self.ncities, i, weight=1)

		self.CompileKernels()


	#Kernel definitions:
	def constrain(self, n=False):
		if n:
			return "constrain"
		return numpy.eye(self.ncities) * self.ConstraintFactor

	def distance(self, n=False):
		if n:
			return "distance"
		return self.distances


	def DisplayState(self, state):
		"""
		If the regular euclidean TSP was used, this plots the city locations
		and connects them with line segments to show the tour represented by the given state.

		:param state: Vector of spin magnetizations.
		:type state: 1-D Numpy int
		"""
		plt.figure()

		#draw cities:
		plt.scatter(self.coords[0,:], self.coords[1,:])

		#draw route:
		for leg in range(self.ncities):
			start_x = self.coords[0,state[leg]]
			start_y = self.coords[1,state[leg]]
			end_x = self.coords[0,state[(leg+1)%self.ncities]]
			end_y = self.coords[1,state[(leg+1)%self.ncities]]
			x = [start_x, end_x]
			y = [start_y, end_y]
			plt.plot(x,y,c='black')

		plt.gca().set_aspect('equal')
		plt.show()

	# def defaultTemp(self, niters, tmax=10):
	# 	PwlTemp = numpy.zeros([2, 3], dtype="float32")
	# 	PwlTemp[0,0] = tmax
	# 	PwlTemp[0,1] = tmax/2
	# 	PwlTemp[0,2] = 0.02
	# 	PwlTemp[1,0] = 0
	# 	PwlTemp[1,1] = niters*0.8
	# 	PwlTemp[1,2] = niters
	# 	self.PwlTemp = PwlTemp


	def StateDict(self, temp):
		import itertools
		#calculates the ground-truth distribution of states for the given temperature.
		#Obviously, only feasible for a small number of cities.
		state = [i for i in range(self.ncities)]
		energies = []
		perms = list(itertools.permutations(state))
		for perm in perms:
			energies.append(self.EvalCost(perm))
		print("There are %i different tours of %i cities"%(len(energies), self.ncities))
		energies = numpy.array(energies)
		probs = numpy.exp(-energies/temp)
		z = numpy.sum(probs)
		probs = probs/z
		print("Avg E: %.3f"%numpy.sum(probs*energies))
		
		#collect energies and states and probs into a dict:
		d = {}
		for st, pr, e in zip(perms, probs, energies):
			d[st] = (pr, e)
		return d


	# I don't think this is needed since valid and invalid sets are distinguished in the main energy calculating code
	# def AvgE(self, temp):
	# 	import itertools
	# 	#calculates the ground-truth distribution of states for the given temperature.
	# 	#Obviously, only feasible for a small number of cities.
	# 	state = [i for i in range(self.ncities)]
	# 	energies = []
	# 	for perm in itertools.permutations(state):
	# 		energies.append(self.EvalCost(perm))
	# 	print("There are %i different tours of %i cities"%(len(energies), self.ncities))
	# 	energies = numpy.array(energies)
	# 	probs = numpy.exp(-energies/temp)
	# 	z = numpy.sum(probs)
	# 	probs = probs/z
	# 	return numpy.sum(probs*energies)

	def IsValidSemantics(self, state):
		#checks if the state is valid, in the sense of whether or not the state follows the basic rules of the combintorial problem
		#valid meaning that the state represents a valid tour of cities,
		#and invalid represents a state that does not correspond to a valid tour of cites (cities visited more than once or not at all)
		if numpy.unique(state).shape[0] == self.ncities:
			return True
		return False



import PottsPlayground.Test.TestBaseTask
import PottsPlayground.Test.TestColoring
import PottsPlayground.Test.TestBinarized
import PottsPlayground.Test.TestAnneal #uses Coloring task for testing, so should come after TestColoring

#remaining task tests try annealing, so should come after TestAnneal
import PottsPlayground.Test.TestTsp


import numpy
import PottsPlayground
from PottsPlayground.Kernels import BinaryQuadratic as BQK

print("Testing: PottsPlayground.Tasks.BaseTask")

#============================basic spin-index model assembly
#manually create an Ising model of a logical AND gate, C = A&B
andgate = PottsPlayground.Tasks.BaseTask()

#three variables, each which take on values of 0 or 1
andgate.SetPartitions([2,2,2])

#connect A to B
andgate.AddKernel(BQK, i=0, j=1, weight=1)
andgate.AddKernel(BQK, i=1, j=0, weight=1)

#connect A to C
andgate.AddKernel(BQK, i=0, j=2, weight=-2)
andgate.AddKernel(BQK, i=2, j=0, weight=-2)

#connect B to C
andgate.AddKernel(BQK, i=2, j=1, weight=-2)
andgate.AddKernel(BQK, i=1, j=2, weight=-2)

andgate.SetBias(0, [0,0])
andgate.SetBias(1, [0,0])
andgate.SetBias(2, [0,3])

andgate.CompileKernels()

#check results:
for A in [0,1]:
	for B in [0,1]:
		for C in [0,1]:
			state = [A, B, C]
			E = andgate.EvalCost(state)
			if (C == (A*B) and E != 0):
				print("FAILED - Energy of valid AND gate is not zero")
			if (A+B == 1 and C == 1 and E != 1):
				print("FAILED - Energy of invalid AND gate should be 1")
			if (A+B == 2 and C == 0 and E != 1):
				print("FAILED - Energy of invalid AND gate should be 1")
			if (A+B == 0 and C == 1 and E != 3):
				print("FAILED - Energy of invalid AND gate should be 3")







#============================compact text-ID based model assembly
#manually create an Ising model of a logical AND gate, C = A&B
andgate = PottsPlayground.Tasks.BaseTask()

#three variables, each which take on values of 0 or 1
andgate.AddSpins([2,2], ['A', 'B'])

andgate.AddWeight(BQK, 'A', 'B', weight=1)
andgate.AddSpins([2], ['C']) #try adding a node after already adding a weight
andgate.AddWeight(BQK, 'A', 'C', weight=-2)
andgate.AddWeight(BQK, 'B', 'C', weight=-2)

andgate.AddBias('C', [0,3]) #other two Spins need no biases

andgate.CompileKernels()

#check results:
for A in [0,1]:
	for B in [0,1]:
		for C in [0,1]:
			state = [A, B, C]
			state = {'A': A, 'B': B, 'C': C}
			E = andgate.EvalCost(state)
			if (C == (A*B) and E != 0):
				print("FAILED - Energy of valid AND gate is not zero")
			if (A+B == 1 and C == 1 and E != 1):
				print("FAILED - Energy of invalid AND gate should be 1")
			if (A+B == 2 and C == 0 and E != 1):
				print("FAILED - Energy of invalid AND gate should be 1")
			if (A+B == 0 and C == 1 and E != 3):
				print("FAILED - Energy of invalid AND gate should be 3")


import PottsPlayground
import random
import numpy

print("Testing: PottsPlayground.Tasks.Binarized")

#test that unconnected graphs work fine
PottsTask = PottsPlayground.Tasks.BaseTask()
PottsTask.AddSpins([2,2,3,3])

def k(n, i, j):
	if n:
		return "rndk-%i-%i"%(i,j)
	else:
		return numpy.random.rand(i,j)

PottsTask.AddWeight(lambda n: k(n, 2, 2), 0, 1, 1.)
PottsTask.AddWeight(lambda n: k(n, 2, 2), 0, 1, 1.) #add it a second time, just to check multiedge functionality
PottsTask.AddWeight(lambda n: k(n, 2, 3), 1, 2, 2.)
PottsTask.AddWeight(lambda n: k(n, 3, 3), 2, 3, 3.)

PottsTask.CompileKernels()

IsingTask = PottsPlayground.Tasks.Binarized(PottsTask)


assert len(IsingTask.ListSpins()) == 8
p_valid, p_invalid = PottsTask.EnergyBands()
i_valid, i_invalid = IsingTask.EnergyBands()
assert len(p_valid) == len(i_valid)

#check energy of a Potts state and equivalent ising state:
pottsstate = numpy.array([1, 0, 2, 1], dtype='int32')
isingstate = IsingTask.PottsToIsingSemantics(pottsstate)

#check that min valid energies are the same
assert (numpy.min(p_valid) + 1e-2 > numpy.min(i_valid))
assert (numpy.min(p_valid) - 1e-2 < numpy.min(i_valid))

#check that invalid energies are all higher than the lowest valid energy:
# assert (numpy.min(i_valid) + 1 - 1e-2 < numpy.min(i_invalid))


"""
Commonly used kernel definitions for constructing Potts models.
Each 'kernel' is a weight matrix between two Potts spins *i* and *j*.
Given a model state, each kernel :math:`W_{i,j}` contributes energy :math:`W_{i,j}[m_i,m_j]`
to the overall model, where :math:`m_i` and :math:`m_j` are the spin values of Potts nodes :math:`i` and :math:`j`.

Each kernel 'definition' is actually a function that returns either a string identifier for the kernel
or the kernel itself.  This is designed so that complex kernels that take a long time to construct are
only constructed once for each problem (A large problem might have thousands of copies of a big kernel that 
is created element by element).  Kernels in this module should work seemlessly with the baseTask AddKernel method,
however many nontrivial applications will require custom-designed kernels.
"""
import numpy

def BinaryQuadratic(n=False):
	"""
	Kernel definition for Binary Quadratic Ising models
	where each spin is interpreted to have values of 0 or 1.
	Is simply

	K = [[0	0]
	     [0	1]]

	Since the processing component of PottsPlayground additionally weighs each kernel,
	the BinaryQuadratic kernel is a fixed value, allowing a simplified Ising computation mode.
	"""
	if n:
		return "BQ"
	k = numpy.zeros([2,2])
	k[1,1] = 1
	return k

def IsingSpin(n=False):
	"""
	Kernel definition for physical Ising models,
	where each spin is interpreted to have values of -1 or +1.
	Is simply

	K = [[-1  1]
	     [ 1 -1]]

	The kernel is designed such that when it is scaled positively,
	It creates correlation between the connected i and j spins;
	and when scaled by a negative value, creates anticorrelation.
	This is to match typical sign conventions in +1/-1 spin systems.
	
	"""
	if n:
		return "IS"
	k = numpy.array([[-1, 1], [1, -1]])
	return k

def Identity(dim=2, n=False):
	"""
	Identity kernel, i.e. all entries are zero except for the diagonal, which have values 1.
	Often used to prevent Potts nodes from being in the same state as each other.

	To use this kernel, the dimension needs to be pre-specified before
	the kernel is passed to AddWeight, such as 

	Identity10 = lambda n: Identity(dim=10, n=n)
	AddWeight(Identity10, spin_i, spin_j)
	
	:param dim: Size of the identity kernel.
	:type dim: int
	"""
	if n:
		return "Identity"
	return numpy.eye(dim)


print("Testing: Examples")


#manually create an Ising model of a logical AND gate, C = A&B

import numpy
import PottsPlayground
from PottsPlayground.Kernels import BinaryQuadratic as BQK

andgate = PottsPlayground.Tasks.BaseTask()

#three variables, each which take on values of 0 or 1
andgate.AddSpins([2,2,2], ['A', 'B', 'C'])

#BQK is the binary quadratic kernel,
#which allows the simplification of the Potts Hamiltonian to the Ising Hamiltonian:
#Potts: E = sum Wij(mi,mj) -> 
#Binary Quadratic Ising: E = sum Wij*mi*mj with m valued as 0 or 1
andgate.AddWeight(BQK, 'A', 'B', weight=1)
andgate.AddWeight(BQK, 'A', 'C', weight=-2)
andgate.AddWeight(BQK, 'B', 'C', weight=-2)

andgate.AddBias('C', [0,3]) #other two Spins need no biases

andgate.CompileKernels()

#run the model at a constant temperature to capture statistics:
temp = PottsPlayground.Schedules.constTemp(niters=1e6, temp=1)
results = PottsPlayground.Anneal(andgate, temp, nReports=int(1e4))

samples = results["AllStates"]
sample_indices = (samples[:,0, andgate.IndexOf('A')]
				+2*samples[:,0,andgate.IndexOf('B')]
				+4*samples[:,0,andgate.IndexOf('C')])
unique, counts = numpy.unique(sample_indices, return_counts=True)

for i, c in zip(unique, counts):
	pr = c/sample_indices.shape[0]
	print("Pr(A=%i, B=%i, C=%i) = %.3f"%(i%2, (i/2)%2, (i/4)%2, pr))



import numpy
from matplotlib import pyplot as plt
from matplotlib.patches import Ellipse
from matplotlib.patches import Rectangle
from matplotlib.legend_handler import HandlerPatch
from matplotlib import cm
import matplotlib
import math
import types
import itertools
import networkx

class BaseTask:
	"""
	Provides a framework for creating, storing, and manipulating Potts models.  Provides a bridge between the semantic structure
	of various combinatorial problems and a Potts-model format used by the Potts annealing engine.  Usually instantiated as a base class,
	but can also be used alone to create arbitrary Potts model tasks.
	"""

	def __init__(self):
		self.InitKernelManager()

	def SetPartitions(self, qSizes):
		#DEPRECIATED interface for specifying the number and size of spins.
		self.AddSpins(qSizes)

	def AddSpins(self, qSizes, names=None):
		"""
		Add Spins to the model by specifying their size (how many states each spin has).
		Spins may be assigned semantic names, otherwise the spins are given an index identifier.

		:param qSizes: A list or numpy array of integers, one for each spin being added.
		:param names: (optional) A mixed list of str or int, by which each of the spins is identified.  Must have the same number of elements as qSizes.
		:return: None
		"""
		qSizes = numpy.array(qSizes, dtype='int32')
		if names is not None:
			for name, q in zip(names, qSizes):
				self.graph.add_node(name, q=q, bias=numpy.zeros([q]))
		else:
			for indx, q in enumerate(qSizes):
				self.graph.add_node(indx, q=q, bias=numpy.zeros([q]))

	def AddBias(self, spin, bias):
		"""
		Modify the bias vector on the given spin, by adding to it's existing value.
		Allows taking a "superpositional" approach to building up the model.

		:param spin: Name or index of the spin to add to.
		:param bias: Numpy array of bias values (one bias for each of the spin's possible states).
		:return: None
		"""
		self.graph.nodes[spin]["bias"] = self.graph.nodes[spin]["bias"] + numpy.array(bias)

	def PinSpin(self, spin, values, weight=100):
		"""
		Restrict a spin to a subset of its possible states.

		For example,
		model.PinSpin("A", [2,3])
		will force spin "A" to be in state 2 or 3.

		:param spin: Name or index of the spin to add to.
		:param values: Spin indexes to keep.  All others will be excluded.
		:type values: list(int)
		:param weight: optional, the magntude of the bias used to pin the spin.
		:return: None
		"""

		self.graph.nodes[spin]["bias"] = self.graph.nodes[spin]["bias"]*0 + weight
		for value in values:
			self.graph.nodes[spin]["bias"][value] = 0

	def SetBias(self, spin, bias):
		"""
		Modify the bias vector on the given spin, by setting it to a new value.

		:param spin: Name or index of the spin to add to.
		:param bias: Numpy array of bias values (one bias for each of the spin's possible states).
		:return: None
		"""
		self.graph.nodes[spin]["bias"] = bias

	def InitKernelManager(self):
		"""
		(Re)Initializes the kernel manager system, which is used for building up weight matrices piece by piece,
		by initializing a few empty variables.

		:return: None
		"""
		self.KernelList = []
		self.KernelDict = {}
		self.graph = networkx.MultiDiGraph()
		self.Names2Indices = {}
		self.Indices2Names = []
		#directed graph, so that the kernel in each direction can be different (usually just a transpose.)
		#MultiGraph, so that multiple kernels can be applied to the same edge.
		
		self.e_th = -1e9

	def AddKernel(self, creator, i, j, weight=1):
		"""
		Adds a directed connection from spin i to spin j.  The connection is a weight matrix unto itself,
		with size [qSizes[i], qSizes[j]].  'creator' is a function that accepts a single boolean arguement;
		when true, it returns a string identifier for the weight kernel between spins i and j.  If false, 
		it returns the actual weight kernel.  This is designed to balance ease of use and efficiency:
		if the same weight kernel is used between many pairs of Potts spins, the weight kernel will only be
		generated once (which may not be trivial depending on how it is generated), but the user does not need
		to keep track of how many times a given weight kernel was used.

		:param creator: A lambda function that takes a single boolean arguement.
		:param i: Name/Index of first spin.
		:type i: str, int
		:param j: Name/Index of second spin.
		:type j: str, int
		:param weight: Scales the entire weight kernel for this connection only.
		:type weight: float
		:return: None
		"""

		#first, see if kernel has already been created, and if not, create it:
		if weight == 0:
			return
		kName = creator(True)
		if kName not in self.KernelDict:
			self.KernelList.append(creator(False))
			self.KernelDict[kName] = len(self.KernelList)-1
		kIndex = self.KernelDict[kName]

		#add the connection to the graph:
		self.graph.add_edge(i, j, weight=weight, kIndex=kIndex)

	def AddWeight(self, creator, i, j, weight=1):
		"""
		Adds an undirected connection between spin i and spin j.  The connection is a weight matrix unto itself,
		with size [qSizes[i], qSizes[j]].  This weight matrix is automatically transposed
		to create the reverse j to i connection.  'creator' is a function that accepts a single boolean arguement;
		when true, it returns a string identifier for the weight kernel between spins i and j.  If false, 
		it returns the actual weight kernel.  This is designed to balance ease of use and efficiency:
		if the same weight kernel is used between many pairs of Potts spins, the weight kernel will only be
		generated once (which may not be trivial depending on how it is generated), but the user does not need
		to keep track of how many times a given weight kernel was used.

		:param creator: A lambda function that takes a single boolean arguement.
		:param i: Name/Index of first spin.
		:type i: str, int
		:param j: Name/Index of second spin.
		:type j: str, int
		:param weight: Scales the entire weight kernel for this connection only.
		:type weight: float
		:return: None
		"""

		#first, see if kernel has already been created, and if not, create it:
		if weight == 0:
			return

		kName = creator(True)
		kNameT = kName + "_transposed"

		if kName not in self.KernelDict:
			self.KernelList.append(creator(False))
			self.KernelDict[kName] = len(self.KernelList)-1
		kIndex = self.KernelDict[kName]

		#mange transposed kernels.
		#there is always a separate named entry for the transposed kernel,
		#but if the kernel is symetric, the name merely points to the same kernel as the non-transposed version.
		if kNameT not in self.KernelDict:
			k = self.KernelList[kIndex]
			kt = numpy.transpose(k)
			if k.shape != kt.shape or numpy.any(k != kt):
				self.KernelList.append(kt)
				self.KernelDict[kNameT] = len(self.KernelList)-1
			else:
				self.KernelDict[kNameT] = kIndex
		kIndexT = self.KernelDict[kNameT]

		#add the connection to the graph:
		self.graph.add_edge(i, j, weight=weight, kIndex=kIndex)
		self.graph.add_edge(j, i, weight=weight, kIndex=kIndexT)

	def Compile(self):
		"""
		After all of the connections in a model have been added via AddKernel, the weight kernels are assembled into
		Numpy arrays with specific formats, tailored for use in the C++/CUDA sampling code. The results are stored in the Task object.

		The weight kernels are stored in a dense format in a 3-D Numpy array.  The first dimension is the kernel index,
		while the second and third dimensions correspond to the magnetization states of the connected spins.

		A second 3-D NUmpy array, the kernel map, specifies which kernel connects any given pair of spins.  It is stored in
		a sparse format.

		:return: None
		"""
		self.CompileKernels() #Renamed from CompileKernels to just Compile, but didn't want to break existing code

	def CompileKernels(self):
		#create mapping from NetworkX graph representation to a list of node sizes.
		#also creates a bidirectional lookup for converting between machine indices and NetworkX names.
		#if the lookup already exists and no changes have been made, leave as-is
		# (in case some weights were changed but the network remains the same).
		self.nnodes = self.graph.number_of_nodes()
		if self.nnodes != len(self.Indices2Names): #if they are the same, make no change to the mapping
			self.Indices2Names = []
			self.Names2Indices = {}
			self.qSizes = numpy.zeros([self.nnodes], dtype='int32')
			for i, node in enumerate(self.graph.nodes):
				self.Indices2Names.append(node)
				self.Names2Indices[node] = i
				self.qSizes[i] = self.graph.nodes[node]['q']

		#build bias matrix:
		self.biases = numpy.zeros([self.nnodes, numpy.max(self.qSizes)], dtype="float32")
		for i, name in enumerate(self.Indices2Names):
			self.biases[i,:self.qSizes[i]] = self.graph.nodes[name]['bias']

		#compile supplementary model metadata for backend use:
		self.qCumulative = numpy.cumsum(self.qSizes, dtype="int32")
		self.Partitions = numpy.zeros([int(self.qCumulative[-1])], dtype="int32")
		self.Partition_states = numpy.zeros([int(self.qCumulative[-1])], dtype="int32")
		for i in range(self.nnodes):
		 	self.Partitions[int(self.qCumulative[i]-self.qSizes[i]):int(self.qCumulative[i])] = i
		 	self.Partition_states[int(self.qCumulative[i]-self.qSizes[i]):int(self.qCumulative[i])] = numpy.linspace(0,self.qSizes[i]-1,int(self.qSizes[i]))


		#dense kernels
		maxQ = numpy.max(self.qSizes)
		nKernels = len(self.KernelList)
		kernels = numpy.zeros([nKernels, maxQ, maxQ], dtype="float16")
		for i, kernel in enumerate(self.KernelList):
			kernels[i, :kernel.shape[0], :kernel.shape[1]] = kernel

		#sparse kmap:
		nPartitions = self.qSizes.shape[0]
		assert nPartitions == self.graph.number_of_nodes()
		maxDensity = numpy.max([self.graph.out_degree(name) for name in self.Names2Indices])
		total_count = 0
		kmap_sparse = numpy.zeros([nPartitions, maxDensity+1,3], dtype="float16")
		for i, name in enumerate(self.Indices2Names): #this could break if there any completely unconnected nodes in the graph.
			kmap_sparse[i,0,0] = self.graph.out_degree(name)+1 #+1, so that it directly tells the stop index, rather than the number of elements
			for c, (u, v, edge_data) in enumerate(self.graph.out_edges(name, data=True)):
				kmap_sparse[i,c+1,:] = [edge_data['kIndex'], edge_data['weight'], self.Names2Indices[v]]
			total_count = total_count + len(self.graph.out_edges(name))
		self.sparse_kmap_density = (total_count/(nPartitions**2))
		# print("Done making sparse kernel map, density = %.3f"%)

		#cast back to float 32.  By starting with float 16 and copying to float32,
		#hopefully the floats will be truncated so as to avoid some cumulative errors in 32 bit FP math
		self.kernels = numpy.zeros([nKernels, maxQ, maxQ], dtype="float32")
		numpy.copyto(self.kernels, kernels)
		self.kmap_sparse = numpy.zeros([nPartitions, maxDensity+1,3], dtype="float32")
		numpy.copyto(self.kmap_sparse, kmap_sparse)


	def ListSpins(self):
		"""
		Returns a list of spin names/identifiers currently in the model.
		"""

		return [node for node in self.graph.nodes]

	def SpinSize(self, spin):
		"""
		Returns the number of states that the given spin can have.
		"""

		return self.graph.nodes[spin]['q']

	def TotalWeight(self, spin_i, spin_j):
		"""
		Returns a numpy array of the weight values between spin i and spin j.
		If there are multiple kernels, they will be added together.
		"""

		w = numpy.zeros([self.SpinSize(spin_i), self.SpinSize(spin_j)])
		for key, edge_data in self.graph.get_edge_data(spin_i, spin_j, default={}).items():
			w = w + edge_data['weight']*self.KernelList[edge_data['kIndex']]

		return w

	def GetSpinBias(self, spin):
		"""
		Returns a numpy vector of the biases acting on the spin.
		"""

		return self.graph.nodes[spin]["bias"]

	def GetSpinFromState(self, spin, state):
		"""
		Given a spin identifier, extract that spin's value from a densely stored state.
		"""
		return state[self.Names2Indices[spin]]

	def SetSpinInState(self, spin, state, value):
		"""
		Given a spin identifier, set that spin's value in a densely stored state.
		"""

		state[self.Names2Indices[spin]] = value 
	
	# ====================================================================================
	#
	def EvalCost(self, state):
		"""
		Python-domain calculation of the energy of a state, based on compiled kernels and sparse kernel map.

		:param state: Magnetization values for each of the spins in the model.
		:type state: 1-D numpy array of ints, or a dict where keys are spin names and values are the spin states.
		:return: Energy/Cost of the model when in the given state.
		:rtype: float
		"""
		if (type(state) == dict):
			state = self.NamedState2IndexState(state)

		cost = 0
		for i,m in enumerate(state):
			cost = cost + 2*self.biases[i,m]
			for c in range(1, int(self.kmap_sparse[i,0,0])):
				j = int(self.kmap_sparse[i,c,2])
				w = self.kmap_sparse[i,c,1]
				k = int(self.kmap_sparse[i,c,0])
				n = int(state[j])
				cost = cost + w*self.kernels[k,m,n]
		return cost/2

	def NamedState2IndexState(self, NamedState):
		IndexedState = []
		for name in self.Indices2Names:
			IndexedState.append(NamedState[name])
		return IndexedState

	def IndexOf(self, spin):
		"""
		Converts from a Spin name used during model construction to 
		the corresponding spin index used for computation.

		:param spin: Spin name used during model construction.
		:return: Spin index in a model state.
		:rtype: int
		"""
		return self.Names2Indices[spin]

	def EvalPE(self, state, i, m):
		"""
		Calculates the potential energy of state m of node i, given state.  

		:param state: Magnetization values for each of the spins in the model.
		:type state: 1-D numpy array of ints.
		:param i: Spin index of interest.
		:type i: int
		:param m: value of spin i
		:type m: int
		:return: The energy that would potentially be added to the total if spin i was in state m.
		:rtype: float
		"""

		PE = self.biases[i,m]
		for c in range(1, int(self.kmap_sparse[i,0,0])):
			j = int(self.kmap_sparse[i,c,2])
			w = self.kmap_sparse[i,c,1]
			k = int(self.kmap_sparse[i,c,0])
			n = int(state[j])
			PE = PE + w*self.kernels[k,m,n]
		return PE

	def EvalDE(self, state, i, m_new):
		"""
		Calculates how much the total energy would change if spin i had magnetization m_new.

		:param state: Magnetization values for each of the spins in the model.
		:type state: 1-D numpy array of ints.
		:param i: Spin index of interest.
		:type i: int
		:param m_new: new value of spin i
		:type m_new: int
		:return: Change in energy if spin i had magnetization m_new.
		:rtype: float
		"""
		m_old = state[i]
		DE = self.biases[i, m_new]-self.biases[i, m_old]
		for c in range(1, int(self.kmap_sparse[i,0,0])):
			j = int(self.kmap_sparse[i,c,2])
			w = self.kmap_sparse[i,c,1]
			k = int(self.kmap_sparse[i,c,0])
			n = int(state[j])
			DE = DE + w*(self.kernels[k,m_new,n]-self.kernels[k,m_old,n])
		return DE

	def EvalDDE(self, i, m_old, m_new):
		"""
		Calculates the delta-delta-energy of all possible magnetizations of all spins.
		DDE is how much the DE of each action changes, when an action is taken.

		:param i: Index of changing spin that creates the DDE.
		:type i: int
		:param m_old: new value of spin i
		:type m_old: int
		:param m_new: new value of spin i
		:type m_new: int
		:return: DDE value for the all magnetizations of all spins.
		:rtype: 2-D Numpy float array, N_{spins}X N_{magnetizations}
		"""
		DDE = numpy.zeros([self.qSizes.shape[0], numpy.max(self.qSizes)])

		for c in range(1, int(self.kmap_sparse[i,0,0])):
			j = int(self.kmap_sparse[i,c,2])
			w = self.kmap_sparse[i,c,1]
			k = int(self.kmap_sparse[i,c,0])
			DDE[j] = DDE[j] + w*(self.kernels[k,m_new,:]-self.kernels[k,m_old,:])

		return DDE


	def EnergyBands(self):
		"""
		Calculates the energy of every possible state.  Since the number of states grows eponentially,
		this can only be used with very small problem instances, and has been used just for illustrative purposes.

		:return valid_energies: A list of energies that correspond to valid model states.  Validity is defined by child classes.
		:return invalid_energies: A list of energies that correspond to invalid model states.
		"""
		state = [i for i in range(numpy.max(self.qSizes))]
		valid_energies = []
		invalid_energies = []
		for perm in itertools.product(state, repeat=self.qSizes.shape[0]):
			#just from how this product method works, we need to check to make sure that 
			#the permutation is a valid state, in case the Potts model has unequal qSizes:
			if not numpy.all(perm < self.qSizes):
				continue #skipping this perm since it is a non-existant state
			if self.IsValidSemantics(perm): #IsValidSemantics is defined by the child class
				valid_energies.append(self.EvalCost(perm))
			else:
				invalid_energies.append(self.EvalCost(perm))
		return valid_energies, invalid_energies

	def IsValidSemantics(self, state):
		return True
		
	#creates a graphic of the weights and states of the model, with semantic annotations
	qannotations = None
	sannotations = None
	sAnnotatePad = 15
	cellgap=0.5
	partspace = 1 #space from the partition units to the weight matrix
	qfont = matplotlib.font_manager.FontProperties(family='Arial', style='normal', size=8, weight='normal', stretch='normal')
	sfont = matplotlib.font_manager.FontProperties(family='Arial', style='normal', size=14, weight='normal', stretch='normal')
	

	def DisplayModel(self, m1=None, m2=None, state=None, flags=[]):
		#m1 and m2 allow subsets of the model to be plotted (otherwise this might be too big of a plot to see anything)
		if m1 is None:
			m1 = [i for i in range(len(self.qSizes))]
		if m2 is None:
			m2 = [i for i in range(len(self.qSizes))]
		qs1 = [self.qSizes[i] for i in m1]
		qs2 = [self.qSizes[i] for i in m2]
		qc1 = numpy.cumsum(qs1, dtype="int32")
		qc2 = numpy.cumsum(qs2, dtype="int32")

		local_org_x = qc1 - qs1 + numpy.linspace(0, self.cellgap*(len(qs1)-1), len(qs1))
		local_org_y = qc2 - qs2 + numpy.linspace(0, self.cellgap*(len(qs2)-1), len(qs2))

		fig = plt.figure()
		gs = fig.add_gridspec(1, 3)#, hspace=0., wspace=0.3)
		ax1 = fig.add_subplot(gs[0, 0:2])
		ax2 = fig.add_subplot(gs[0, 2])

		plt.sca(ax1)
		ax1 = plt.gca()
		ymax = local_org_y[-1]+qs2[-1]+0.5
		xmax = local_org_x[-1]+qs1[-1]+0.5
		# plt.gca().add_patch(Rectangle((-0.5, -0.5), qc1[-1]+0.5+0.5*(len(qs1)), qc2[-1]+0.5+0.5*(len(qs2)), fc='lightgray', zorder=0))

		#build the weight matrix:
		weights = numpy.zeros([qc1[-1], qc2[-1]], dtype="float32")
		for ii, i in enumerate(m1):
			if not self.graph.has_node(i):# not in self.kMapLists:
				continue
			for u, v, edge_data in self.graph.out_edges(i, data=True):
				
				#start by getting dimensions and positions of the particular kernel:
				j = v
				k = edge_data['kIndex']
				if j not in m2:
					continue
				jj = m2.index(j)
				sz_x = int(self.qSizes[i])
				sz_y = int(self.qSizes[j])
				pos_x = int(qc1[ii]-sz_x)
				pos_y = int(qc2[jj]-sz_y)
				k = self.kernels[k, 0:sz_x, 0:sz_y]
				
				weights[pos_x:pos_x+sz_x, pos_y:pos_y+sz_y] = weights[pos_x:pos_x+sz_x, pos_y:pos_y+sz_y] + k*edge_data['weight']

		#if a state is available, also calculate PE and DE values
		if state is not None:
			pe_vals = numpy.zeros([len(m1), numpy.max(qs1)])
			de_vals = numpy.zeros([len(m1), numpy.max(qs1)])
			for ii, i in enumerate(m1):
				for q in range(qs1[ii]):
					pe_vals[ii, q] = self.EvalPE(state, i, q)
					de_vals[ii, q] = self.EvalDE(state, i, q)

		
		#set up colorbars to indicate weight meaning
		def custom_cmap(positive_color, negative_color=None):
			zero_color = 'lightgrey'
			if negative_color is not None:
				return matplotlib.colors.LinearSegmentedColormap.from_list('NA name', [negative_color, zero_color, positive_color])
			else:
				return matplotlib.colors.LinearSegmentedColormap.from_list('NA name', [zero_color, positive_color])

		cmapWeights = custom_cmap('black')
		normWeights = matplotlib.colors.Normalize(vmin=numpy.min(weights), vmax=numpy.max(weights))
		cmapWeights = matplotlib.cm.ScalarMappable(norm=normWeights, cmap=cmapWeights)

		cmapActiveWeights = custom_cmap('salmon')
		cmapActiveWeights = matplotlib.cm.ScalarMappable(norm=normWeights, cmap=cmapActiveWeights)

		if 'DE' in flags:
			cmapEnergy = custom_cmap('green', 'blue')
			absmax = numpy.max([numpy.max(numpy.abs(pe_vals)), numpy.max(numpy.abs(de_vals))])
			normEnergy = matplotlib.colors.Normalize(vmin=-absmax, vmax=absmax)
		elif 'PE' in flags:
			cmapEnergy = custom_cmap('green')
			normEnergy = matplotlib.colors.Normalize(vmin=-numpy.min(pe_vals), vmax=numpy.max(pe_vals))
			cmapEnergy = matplotlib.cm.ScalarMappable(norm=normEnergy, cmap=cmapEnergy)


		#separate back out into kernels for plotting:
		for ii, i in enumerate(m1):
			for jj, j in enumerate(m2):
				sz_x = int(self.qSizes[i])
				sz_y = int(self.qSizes[j])
				pos_x = int(qc1[ii]-sz_x)
				pos_y = int(qc2[jj]-sz_y)
				k = weights[pos_x:pos_x+sz_x, pos_y:pos_y+sz_y]
				img = cmapWeights.to_rgba(k)
				if state is not None and 'PE' in flags:
					img[:, state[j], :] = cmapActiveWeights.to_rgba(k[:, state[j]])
				elif state is not None and 'no-hl' not in flags:
					img[state[i], state[j], :] = cmapActiveWeights.to_rgba(k[state[i], state[j]])
				img = numpy.transpose(img, axes=[1,0,2])

				x_ofst = local_org_x[ii]
				y_ofst = local_org_y[jj]
				plt.imshow(img, extent=[x_ofst, x_ofst+sz_x, y_ofst+sz_y, y_ofst])

		plt.colorbar(mappable=cmapWeights, ax=ax2, orientation='horizontal', label='weight strength', aspect=10)
		plt.colorbar(mappable=cmapActiveWeights, ax=ax2, orientation='horizontal', label='active weight strength', aspect=10)
		if 'PE' in flags or 'DE' in flags:
			plt.colorbar(mappable=cmapEnergy, ax=ax2, orientation='horizontal', label='partial energy', aspect=10)

		#making the legend.  Oh my god why am I writing so much complicated code just for these stupid plots...
		class HandlerEllipse(HandlerPatch):
			def create_artists(self, legend, orig_handle,
							   xdescent, ydescent, width, height, fontsize, trans):
				center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
				p = Ellipse(xy=center, width=height + xdescent,
									 height=height + ydescent)
				self.update_prop(p, orig_handle, legend)
				p.set_transform(trans)
				return [p]

		state_option_artist = Ellipse((0, 0), 0.7, 0.7, fc='gray')
		state_active_artist = Ellipse((0, 0), 0.7, 0.7, fc='salmon')
		ax2.legend([state_option_artist, state_active_artist], ['Magnetization Option', 'Current Magnetization'], handler_map={Ellipse: HandlerEllipse()})
		ax2.spines[['right', 'top', 'left', 'bottom']].set_visible(False)
		ax2.set_xticks([])
		ax2.set_yticks([])

		#========================================================================================================================
		#start adding partition embelishments.
		topspace = 0
		leftspace = 0

		if 'PE' in flags or 'DE' in flags:
			topspace = topspace + 1
			for ii, i in enumerate(m1):
				for q in range(qs1[ii]):
					flt_indx = local_org_x[ii] + q + 0.5
					plt.arrow(flt_indx, 0, 0, -1, length_includes_head=True, head_width=0.4, head_length=0.4, color='salmon')

		if 'PE' in flags:
			topspace = topspace + 1
			for ii, i in enumerate(m1):
				for q in range(qs1[ii]):
					flt_indx = local_org_x[ii] + q + 0.5
					color = cmapEnergy.to_rgba(pe_vals[i,q])
					plt.gca().add_patch(Rectangle((flt_indx-0.5, -topspace), 1, 1, fc=color))

		if 'DE' in flags:
			topspace = topspace + 1.3
			for ii, i in enumerate(m1):
				for q in range(qs1[ii]):
					flt_indx = local_org_x[ii] + q + 0.5
					color = cmapEnergy.to_rgba(de_vals[i,q])
					plt.gca().add_patch(Rectangle((flt_indx-0.5, -topspace), 1, 1, fc=color))



		if state is not None:
			topspace = topspace + 1
			leftspace = leftspace + 1

			#add circles to indicate state options/actual state.
			for ii, i in enumerate(m2):
				#but first, add a partition group indicator, using a rectangle and circles at the ends:
				center = local_org_y[ii]+0.5*qs2[ii]
				h = qs2[ii]-1
				plt.gca().add_patch(Ellipse((-leftspace, center+h/2), 0.9, 0.9, fc='lightgray'))
				plt.gca().add_patch(Ellipse((-leftspace, center-h/2), 0.9, 0.9, fc='lightgray'))
				plt.gca().add_patch(Rectangle((-leftspace-0.45, center-h/2), 0.9, h, fc='lightgray'))
				for q in range(qs2[ii]):
					flt_indx = local_org_y[ii] + q + 0.5
					if q == state[i]:
						plt.gca().add_patch(Ellipse((-leftspace, flt_indx), 0.7, 0.7, fc='salmon'))
						if 'no-hl' in flags:
							continue
						if 'PE' in flags: 
							#different visual styles to distinguish potential energy diagrams from total energy diagrams
							plt.arrow(-leftspace, flt_indx, leftspace, 0, length_includes_head=True, head_width=0.4, head_length=0.4, color='salmon')
							plt.plot([0, 0, xmax-0.5, xmax-0.5, 0], [flt_indx-0.5, flt_indx+0.5, flt_indx+0.5, flt_indx-0.5, flt_indx-0.5], c='salmon')
						else:
							plt.plot([-leftspace, xmax], [flt_indx, flt_indx], c='salmon')
					else:
						plt.gca().add_patch(Ellipse((-leftspace, flt_indx), 0.7, 0.7, fc='gray'))


			for ii, i in enumerate(m1):
				#but first, add a partition group indicator, using a rectangle and circles at the ends:
				center = local_org_x[ii]+0.5*qs1[ii]
				h = qs1[ii]-1
				plt.gca().add_patch(Ellipse((center+h/2, -topspace), 0.9, 0.9, fc='lightgray'))
				plt.gca().add_patch(Ellipse((center-h/2, -topspace), 0.9, 0.9, fc='lightgray'))
				plt.gca().add_patch(Rectangle((center-h/2, -topspace-0.45), h, 0.9, fc='lightgray'))
				for q in range(qs1[ii]):
					flt_indx = local_org_x[ii] + q + 0.5
					if q == state[i]:
						plt.gca().add_patch(Ellipse((flt_indx, -topspace), 0.7, 0.7, fc='salmon'))
						if 'no-hl' in flags:
							continue
						if 'PE' not in flags:
							plt.plot([flt_indx, flt_indx], [xmax, -topspace], c='salmon')
					else:
						plt.gca().add_patch(Ellipse((flt_indx, -topspace), 0.7, 0.7, fc='gray'))



		if self.qannotations is not None:
			xlbls = []
			xlblpos = []
			for ii, i in enumerate(m1):
				for q in range(qs1[ii]):
					flt_indx = local_org_x[ii] + q + 0.5
					xlblpos.append(flt_indx)
					xlbls.append(self.qannotations[i][q])
			plt.xticks(xlblpos, xlbls)
			
			ylbls = []
			ylblpos = []
			for ii, i in enumerate(m2):
				for q in range(qs2[ii]):
					flt_indx = local_org_y[ii] + q + 0.5
					ylblpos.append(flt_indx)
					ylbls.append(self.qannotations[i][q])
			plt.yticks(ylblpos, ylbls)
		else:
			plt.xticks([])
			plt.yticks([])

		axx2 = ax1.secondary_xaxis('top')
		axy2 = ax1.secondary_yaxis('left')
		if self.sannotations is not None:
			pos = [local_org_x[i] + 0.5*qs1[i] for i in range(len(qs1))]
			ann = [self.sannotations[i] for i in m1]
			axx2.set_xticks(pos, ann)
			pos = [local_org_y[i] + 0.5*qs2[i] for i in range(len(qs2))]
			ann = [self.sannotations[i] for i in m2]
			axy2.set_yticks(pos, ann)
		

		ax1.tick_params(top=True, labeltop=True, bottom=False, labelbottom=False)
		

		for ax, font in zip([ax1, axx2, axy2], [self.qfont, self.sfont, self.sfont]):
			ax.spines[['right', 'top', 'left', 'bottom']].set_visible(False)
			for label in ax.get_xticklabels() :
				label.set_fontproperties(font)
			for label in ax.get_yticklabels() :
				label.set_fontproperties(font)

		
		axx2.tick_params(axis='x', pad=self.sAnnotatePad, length=0)
		axy2.tick_params(axis='y', pad=self.sAnnotatePad, length=0)
		plt.setp(axx2.xaxis.get_majorticklabels(), rotation=45, ha="left", rotation_mode="anchor")
		plt.setp(ax1.xaxis.get_majorticklabels(), rotation=45, ha="left", rotation_mode="anchor") 
		# for ax in [axx2, axy2]:


		for ax in [ax1, axx2]:
			ax.set_xlim([-leftspace-0.5, xmax])
		for ax in [ax1, axy2]:
			ax.set_ylim([ymax, -topspace-0.5])

		plt.tight_layout()
		plt.show()




	# ================================================================================ Potts kernel/weight construction utilities
	def DisplayKernels(self):
		for name in self.KernelDict:
			plt.figure() #show each different kernel sequentially
			k = -self.KernelList[self.KernelDict[name]] #negate, so darker means stronger weight
			print("Kernel %s has dimensions %i,%i"%(name, k.shape[0], k.shape[1]))
			plt.imshow(k, cmap='gray', vmin=numpy.min(k), vmax=numpy.max(k))
			plt.title(name)
			plt.show()

	def PrintCombinatorics(self):
		nstates = 1.
		for q in self.qSizes:
			nstates = nstates*q
		print("Total number of states:", nstates)


import numpy
from PottsPlayground.Tasks.BaseTask import BaseTask
from PottsPlayground.Kernels import BinaryQuadratic as BQK
from matplotlib import pyplot as plt

class Binarized(BaseTask):
	"""
	Maps any Potts model to an Ising model by splitting every q-dimensional Spin
	into q 2-dimensional spins and adding soft one-hot constraints 
	to each group of Ising spins.
	"""
	def __init__(self, PottsTask, ConstraintFactor=1):
		"""
		Turns a Potts model into an Ising model.  The resulting task is still a PottsTask, but all spins are binary.  The task can still
		be annealed as a Potts model, but can additionally be annealed using dedicated Ising-model code that runs a little faster.

		:param PottsTask: Any object correctly derived from the BaseTask class. Could even be another BinarizedTask, although that would not be productive.
		:param ConstraintFactor: Weight penalty for two Ising bits both being "hot" if they represent different values of the same Potts spin.
		:type ConstraintFactor: float
		"""
		BaseTask.__init__(self)
		self.PottsTask = PottsTask
		self.e_th = PottsTask.e_th

		PottsSpins = PottsTask.ListSpins()

		#create new partitions and set biases
		for i, spin in enumerate(PottsSpins):
			q = PottsTask.SpinSize(spin)
			biases = PottsTask.GetSpinBias(spin)
			subspins = ["%i-%i"%(i,j) for j in range(q)]
			if q == 2:
				self.AddSpins([2], [subspins[0]])
				self.SetBias(subspins[0], biases)
			else:
				self.AddSpins([2]*q, subspins)
				[self.SetBias(subspins[j], [0,biases[j]]) for j in range(q)]

				#add one-hot inhibitory weights.
				#zero bias, one bias, and correlation penalty for enforcing one-hotness amoung Pbits in a Potts node.
				#Designed so that the energy is zero when the collection of P-bits is one-hot,
				#energy is E+=ConstraintFactor if there are two hot or zero hot, and E += >ConstraintFactor if more than two are hot.
				H0 = ConstraintFactor/q 
				H1 = -(q-1)*H0
				w_inhib = 2*q*H0
				for j in range(q):
					self.AddBias(subspins[j], [H0, H1])
					for jj in range(j+1, q):
						self.AddWeight(BQK, subspins[j], subspins[jj], weight=w_inhib)

		for i, spin1 in enumerate(PottsSpins):
			for j, spin2 in enumerate(PottsSpins):
				if j <= i:
					continue #assume symetric, only iterate over half
				w = PottsTask.TotalWeight(spin1, spin2)
				if numpy.sum(w) == 0:
					continue #no weight here

				#if either Potts spin is only size 2,
				#simplify it to a weird sort of sudo-size-1 potts spin
				#by factoring some of the weights into the biases,
				#and then the remaining size-1 weight matrix can be processed
				#just like any size q weight matrix
				if w.shape[0] == 2:
					w[1,:] = w[1,:]-w[0,:]
					if w.shape[1] == 2:
						self.AddBias("%i-%i"%(j,0), w[0,:])
					else:
						[self.AddBias("%i-%i"%(j,y), [0, w[0,y]]) for y in range(w.shape[1])]
					w = w[1,:]
					w = numpy.expand_dims(w, 0)

				if w.shape[1] == 2:
					w[:,1] = w[:,1]-w[:,0]
					if w.shape[0] == 1:
						#this case is slightly different than above, 
						#since w would have been modified above
						self.AddBias("%i-%i"%(i,0), [0,w[0,0]])
					else:
						[self.AddBias("%i-%i"%(i,x), [0, w[x,0]]) for x in range(w.shape[0])]
					w = w[:,1]
					w = numpy.expand_dims(w, 1)

				q1, q2 = w.shape
				for x in range(q1): #iterate through all the values in the kernel:
					for y in range(q2):
						self.AddWeight(BQK, "%i-%i"%(i,x), "%i-%i"%(j,y), weight=w[x,y])


		# self.SetPartitions(numpy.array([2]*numpy.sum(PottsTask.qSizes)))

		# #copy biases first:
		# for Pi, (q, qc) in enumerate(zip(PottsTask.qSizes, PottsTask.qCumulative)):
		# 	m10 = qc-q #the P-bit number corresponding to the first state of Potts node i
		# 	for Pmi in range(q): #Potts m_i value
		# 		Ii = qc-q+Pmi #Ising i index
		# 		self.AddBias(Ii, [0,PottsTask.biases[Pi, Pmi]])
				# self.biases[Ii, 1] = PottsTask.biases[Pi, Pmi]

		#iterate through the Potts Model's Kernel system,
		#copying weights over into ising model Kernel system
		# self.InitKernelManager()
		# nUnits = PottsTask.qSizes.shape[0]
		# for i in range(nUnits): #iterates through rows of weight matrix
		# 	for c, (i, j, edge_data) in enumerate(PottsTask.graph.out_edges([i], data=True)):
		# 		# kmap_sparse[i,c+1,:] = [edge_data['kIndex'], edge_data['weight'], v]
		# 	# for kspec in PottsTask.kMapLists[i]: #iterates through sparse format of column kernels
		# 		#start by getting dimensions and positions of the particular kernel:
		# 		# j = v#kspec[2]
		# 		k = edge_data['kIndex']#kspec[0]
		# 		w = edge_data['weight']
		# 		sz_x = int(PottsTask.qSizes[i])
		# 		sz_y = int(PottsTask.qSizes[j])
		# 		pos_x = int(PottsTask.qCumulative[i]-sz_x)
		# 		pos_y = int(PottsTask.qCumulative[j]-sz_y)
		# 		k = PottsTask.kernels[k, 0:sz_x, 0:sz_y]*w
		# 		for x in range(sz_x): #iterate through all the values in the kernel:
		# 			for y in range(sz_y):
		# 				self.AddKernel(BQK, pos_x+x, pos_y+y, weight=k[x,y])

		#add additional kernels for representing the one-hot inhibition within Potts nodes:
		# for i, (q, qc) in enumerate(zip(PottsTask.qSizes, PottsTask.qCumulative)):
			
		self.PottsSpins = PottsSpins
		self.CompileKernels()


	def IsingToPottsSemantics(self, IsingState):
		"""
		if the Ising state is valid, this will return the Potts state for the associated Potts Model.
		Invalid states will also be processed into a valid Potts state, but the transformation is undefined.
		The returned Potts state can then be sent to the parent Potts Task for interpretation or visualization as a graph coloring solution, etc.

		:param IsingState: Vector of Ising bit states.
		:type IsingState: 1-D numpy int
		:return: A vector representing a Potts model state.
		:rtype: 1-D numpy int
		"""
		PottsState = numpy.zeros([len(self.PottsSpins)], dtype='int32')
		for i, PottsSpin in enumerate(self.PottsSpins):
			q = self.PottsTask.SpinSize(spin)
			subspins = ["%i-%i"%(i,j) for j in range(q)]
			if q == 2:
				s = self.GetSpinFromState(subspins[0], IsingState)
				self.PottsTask.SetSpinInState(PottsSpin, PottsState, s)
			else:
				s = [self.GetSpinFromState(subspins[j], IsingState) for j in range(q)]
				self.PottsTask.SetSpinInState(PottsSpin, PottsState, numpy.argmax(s))
		return PottsState

		# for i, (q, qc) in enumerate(zip(self.PottsTask.qSizes, self.PottsTask.qCumulative)):
		# 	Potts_node_pbits = IsingState[qc-q:qc]
		# 	PottsState.append(numpy.argmax(Potts_node_pbits))
		# return numpy.array(PottsState, dtype='int32')

	def PottsToIsingSemantics(self, PottsState):
		"""
		Converts a Potts state from the source task into an eqivalent Ising state.

		:param PottsState: Vector of Potts spin states.
		:type PottsState: 1-D numpy int
		:return: A vector representing a Ising model state.
		:rtype: 1-D numpy int
		"""
		IsingState = numpy.zeros([len(self.ListSpins())], dtype='int32')
		for i, PottsSpin in enumerate(self.PottsSpins):
			s = self.PottsTask.GetSpinFromState(PottsSpin, PottsState)
			q = self.PottsTask.SpinSize(PottsSpin)
			subspins = ["%i-%i"%(i,j) for j in range(q)]
			if q == 2:
				self.SetSpinInState("%i-%i"%(i,0), IsingState, s)
			else:
				self.SetSpinInState("%i-%i"%(i,s), IsingState, 1)
		return IsingState

		# for i, q in enumerate(self.PottsTask.qSizes):
		# 	partial_state = [0]*q
		# 	partial_state[PottsState[i]] = 1
		# 	IsingState = IsingState + partial_state
		# return numpy.array(IsingState, dtype='int32')

	def IsValidSemantics(self, IsingState):
		#checks if the state is valid, in the sense of whether or not the state follows the basic rules of the combintorial problem
		#valid meaning that the state represents a valid tour of cities,
		#and invalid represents a state that does not correspond to a valid tour of cites (cities visited more than once or not at all)
		
		# for i, (q, qc) in enumerate(zip(self.PottsTask.qSizes, self.PottsTask.qCumulative)):
		# 	Potts_node_pbits = state[qc-q:qc]
		# 	if numpy.sum(Potts_node_pbits) != 1:
		# 		return False

		PottsState = numpy.zeros([len(self.PottsSpins)], dtype='int32')
		for i, PottsSpin in enumerate(self.PottsSpins):
			q = self.PottsTask.SpinSize(PottsSpin)
			subspins = ["%i-%i"%(i,j) for j in range(q)]
			if q == 2:
				s = self.GetSpinFromState(subspins[0], IsingState)
				self.PottsTask.SetSpinInState(PottsSpin, PottsState, s)
			else:
				s = [self.GetSpinFromState(subspins[j], IsingState) for j in range(q)]
				if numpy.sum(s) != 1:
					return False
				self.PottsTask.SetSpinInState(PottsSpin, PottsState, numpy.argmax(s))
		
		#finally, also check if the state is valid in the Potts model sense
		return self.PottsTask.IsValidSemantics(PottsState)



from PottsPlayground.Tasks.Binarized import Binarized
from PottsPlayground.Tasks.MinorEmbedded import MinorEmbedded
from PottsPlayground.Tasks.TravelingSalesman import TravelingSalesman
from PottsPlayground.Tasks.GraphColoring import GraphColoring
from PottsPlayground.Tasks.Ice40Placer import Ice40Placer
from PottsPlayground.Tasks.BaseTask import BaseTask


import PottsPlayground
import random
import numpy

print("Testing: PottsPlayground.Tasks.GraphColoring")

#test that unconnected graphs work fine
PottsTask = PottsPlayground.Tasks.GraphColoring(ncolors=3, p=0, nnodes=4)
valid_energies, invalid_energies = PottsTask.EnergyBands()
assert numpy.sum(valid_energies) == 0

PottsTask = PottsPlayground.Tasks.GraphColoring(ncolors=1, p=1.0, nnodes=4)
valid_energies, invalid_energies = PottsTask.EnergyBands()
assert numpy.sum(valid_energies) == 6

PottsTask = PottsPlayground.Tasks.GraphColoring(ncolors=4, p=1.0, nnodes=4)
valid_energies, invalid_energies = PottsTask.EnergyBands()
assert len(valid_energies) == 4**4
assert numpy.sum(numpy.array(valid_energies) == 0) == 4*3*2*1 


import setuptools
import numpy
import shutil

with open("PottsPlayground/__init__.py", "r") as f:
    for line in f:
        if line.startswith("__version__"):
            # Extract the version string (e.g., "__version__ = '0.1.0'")
            version = line.split("=")[1].strip().strip('"').strip("'")
            break

#if NVCC is found, build with GPU support, else just build for CPU:
if shutil.which("nvcc") is not None:

    from setuptools_cuda_cpp import CUDAExtension, BuildExtension, fix_dll #builds using NVCC

    build={'build_ext': BuildExtension}
    ext = CUDAExtension(
            name="PottsPlayground.Annealing",
            sources=(
                "PottsPlayground/Solvers/Annealing.cpp",
                "PottsPlayground/Solvers/CpuCore.cpp",

                "PottsPlayground/Solvers/GpuCore.cu"
                ),
            include_dirs=[numpy.get_include()],
            libraries=["cudart"],

            extra_compile_args={"cxx": ["-std=c++11"], "nvcc": ["-std=c++11"]},
        )


else:
    build={}
    ext = setuptools.Extension(
            name="PottsPlayground.Annealing",
            sources=[
                "PottsPlayground/Solvers/Annealing.cpp",
                "PottsPlayground/Solvers/CpuCore.cpp",

                "PottsPlayground/Solvers/GpuCoreAlt.cpp"
                ],
            include_dirs=[numpy.get_include()],

            # extra_compile_args=["-std=c++17"]#, "/std:c++11"], #first one is for linuxland, second for stupid MSVC
        )

setuptools.setup(
    name="PottsPlayground",
    version=version,
    packages=['PottsPlayground', 'PottsPlayground.Test', 'PottsPlayground.Tasks'],
    package_dir = { "PottsPlayground":          "PottsPlayground",
                    "PottsPlayground.Test":     "PottsPlayground/Test",
                    "PottsPlayground.Tasks":    "PottsPlayground/Tasks"},
    install_requires = ["matplotlib>=0.0.0",
                        "networkx>=2.0",
                        "numpy>=1.17.0",
                        #"blifparser>=0.0.0",
                        #"minorminer>=0.0.0",
                        #"dimod>=0.0.0",
                        #"dwave-system>=0.0.0"
                        ],
    ext_modules=[ext],
    cmdclass=build,
)

#good references for how to use setuptools:
# https://setuptools.pypa.io/en/latest/userguide/index.html
# https://packaging.python.org/en/latest/guides/modernize-setup-py-project/#modernize-setup-py-project


"""
A collection of definitions for annealing schedules.  Each function returns a 2-D numpy array
that can be interpreted as a piecewise linear function inside of PottsPlayground.Anneal.

The piecewise linear function defines multiple aspects of an annealing run: the number of iterations/spin updates,
a changing temperature as annealing progresses, and at what points annealing should be temporarily halted
so that intermediate data can be recorded.  The intermediate points are the discontinuities in the Piecewise linear function.
Note that progress can also be recorded more frequently, on a set periodic frequency, using the 
nReports option of the PottsPlayground.Anneal function.
"""
import numpy

def constTemp(niters, temp):
	"""
	Anneal a model at a constant temperature for the given number of iterations.

	:param niters: The total number of updates/spin flips over the course of the annealing run.
	:type niters: int
	:param temp: Computational temperature at which the model operates.
	:type temp: float
	"""
	PwlTemp = numpy.zeros([2, 2], dtype="float32")
	PwlTemp[0,0] = temp
	PwlTemp[0,1] = temp
	PwlTemp[1,0] = 0
	PwlTemp[1,1] = niters
	return PwlTemp

def LinearTemp(niters, temp, temp2=0.0001):
	"""
	Anneal a model at a linearly changing temperature, starting at temp and ending at temp2.

	:param niters: The total number of updates/spin flips over the course of the annealing run.
	:type niters: int
	:param temp: Initial temperature.
	:type temp: float
	:param temp2: Final temperature.
	:type temp2: float
	"""
	PwlTemp = numpy.zeros([2, 2], dtype="float32")
	PwlTemp[0,0] = temp
	PwlTemp[0,1] = temp2
	PwlTemp[1,0] = 0
	PwlTemp[1,1] = niters
	return PwlTemp

def constTempLog2Space(niters, temp, nReports):
	"""
	Anneal a model at a constant temperature for the given number of iterations.
	Defines nReports breakpoints spaced logarithmically, 
	so that both fast and slow dynamics of model evolution can be recorded.

	:param niters: The total number of updates/spin flips over the course of the annealing run.
	:type niters: int
	:param temp: Constant temperature used during sampling.
	:type temp: float
	:param nReports: The number of times over the course of annealng that the model is stopped and measured.  The duration ratio between samples is calculated to meet this spec.
	:type nReports: int
	"""
	PwlTemp = numpy.zeros([2, nReports+1], dtype="float32")
	PwlTemp[0,:] = temp
	ratio = niters**(1./(nReports-1))
	for i in range(nReports):
		PwlTemp[1,i+1] = int((ratio**i))
		if (PwlTemp[1,i+1] <= PwlTemp[1,i]):
			PwlTemp[1,i+1] = PwlTemp[1,i] + 1 #just for very begining, so that each step is at least 1 iter more than the last
	PwlTemp[1,0] = 0
	return PwlTemp

def SawtoothTempLog2Space(niters, temp, MinTemp, nTeeth, FirstTooth=10):
	"""
	Anneal a model according to a sawtooth temperature waveform.
	The duration of each sawtooth increases exponentially.
	Unless nReports is defined in the annealing run, a sample will be recorded at the end of each
	sawtooth when the model has been frozen into at least a local minima.
	Allows the "problem solving" capacity of an annealing setup to be observed more directly,
	especially for models that only mix at high temperatures but only settle into valid states at low temperatures.

	:param niters: The total number of updates/spin flips over the course of the annealing run.
	:type niters: int
	:param temp: Initial, high temperature of the sawtooth temperature profile.
	:type temp: float
	:param MinTemp: The temperature at the end of each linearly changing sawtooth cycle.
	:type MinTemp: float
	:param nTeeth: The number of sawtooth cycles over the course of the annealing run.  The duration ratio between teeth is calculated to meet this spec.
	:type nTeeth: int
	:param FirstTooth: How many iterations the first tooth should be, or equivalently.
	:type FirstTooth: int
	"""
	PwlTemp = numpy.zeros([2, nTeeth*2+1], dtype="float32")
	ratio = (niters/FirstTooth)**(1./(nTeeth-1))
	for i in range(nTeeth):
		PwlTemp[1,i*2+1] = int(FirstTooth*(ratio**i))   #bottom of tooth
		PwlTemp[1,i*2+2] = int(FirstTooth*(ratio**i))+1 #top of next tooth
		PwlTemp[0,i*2+1] = MinTemp   #bottom of tooth
		PwlTemp[0,i*2+2] = temp #top of next tooth
		# if (PwlTemp[1,i*2+1] <= PwlTemp[1,i]):
			# PwlTemp[1,i*2+1] = PwlTemp[1,i*] + 1 #just for very begining, so that each step is at least 1 iter more than the last
	PwlTemp[1,0] = 0
	PwlTemp[0,0] = temp
	return PwlTemp